# ======================================================================
#
# Cosmograil: cosmograil.tools.sexcatalog
#
# sexcatalog module.
#
# Author: Laurent Le Guillou <laurentl@ster.kuleuven.ac.be>
#
# $Id: sexcatalog.py,v 1.1 2005/06/29 13:07:41 hack Exp $
#
# ======================================================================
#
# "sexcatalog": python module to read and parse SExtractor catalogs
# A simple interface to read SExtractor text catalogs
#
# ======================================================================
#
# $Log: sexcatalog.py,v $
# Revision 1.1  2005/06/29 13:07:41  hack
# Added Python interface to SExtractor to STSDAS$Python
# for use with 'tweakshifts'. WJH
#
# Revision 1.9  2005/02/14 19:27:31  laurentl
# Added write facilities to rdb module.
#
# Revision 1.8  2005/02/14 17:47:02  laurentl
# Added iterator interface
#
# Revision 1.7  2005/02/14 17:16:30  laurentl
# clean now removes the NNW config file too.
#
# Revision 1.2  2005/02/14 17:13:49  laurentl
# *** empty log message ***
#
# Revision 1.1  2005/02/14 11:34:10  laurentl
# quality monitor now uses SExtractor wrapper.
#
# Revision 1.5  2005/02/11 14:40:35  laurentl
# minor changes
#
# Revision 1.4  2005/02/10 20:15:14  laurentl
# Improved SExtractor wrapper.
#
# Revision 1.2  2005/02/09 23:32:50  laurentl
# Implemented SExtractor wrapper
#
# Revision 1.1  2005/01/06 12:29:25  laurentl
# Added a SExtractor wrapper module. Renamed sextractor.py sexcatalog.py.
#
# Revision 1.1  2004/12/09 03:06:23  laurentl
# Changed tree structure
#
# Revision 1.5  2004/11/26 18:26:59  laurentl
# Added a module to manage the data tree.
#
# Revision 1.4  2004/11/24 15:11:31  laurentl
# Fixed a lot of bugs in sexcatalog module.
#
# Revision 1.2  2004/11/23 22:38:23  laurentl
# Added sexcatalog module.
#
#
# ======================================================================


"""
A simple interface to manipulate SExtractor ASCII catalogs

A simple interface to manipulate SExtractor ASCII catalogs
through a file-like API (open, read, readline, etc.).
For the moment only reading ('r' mode) is supported.
by Laurent Le Guillou
version: 0.1.5 - last modified: 2005-02-14

Future: implement a 'w' mode to be able to save catalogs
in SExtractor format.

Examples:

-----------------------------------------------------------------

    # Through sexcatalog module
    import sexcatalog

    # Read a SExtractor ASCII catalog

    # First method: read the whole catalog at once
    catalog_f = sexcatalog.open(catalog_name)
    catalog = catalog_f.readlines()
    for star in catalog:
        print star['FLUX_BEST'], star['FLAGS']
        if (star['FLAGS'] & sexcatalog.BLENDED):
            print "This star is BLENDED"
    catalog_f.close()

    # Second method: read the catalog star by star
    catalog_f = sexcatalog.open(catalog_name)
    for star in catalog_f:
        print star['FLUX_BEST'], star['FLAGS']
        if (star['FLAGS'] & sexcatalog.BLENDED):
            print "This star is BLENDED"
    catalog_f.close()

    # -------------

    # Through sextractor module
    import sextractor

    # Read a SExtractor ASCII catalog

    # First method: read the whole catalog at once
    catalog_f = sextractor.open(catalog_name)
    catalog = catalog_f.readlines()
    for star in catalog:
        print star['FLUX_BEST'], star['FLAGS']
        if (star['FLAGS'] & sextractor.BLENDED):
            print "This star is BLENDED"
    catalog_f.close()

    # Second method: read the catalog star by star
    catalog_f = sextractor.open(catalog_name)
    star = catalog_f.readline()
    while star:
        print star['FLUX_BEST'], star['FLAGS']
        if (star['FLAGS'] & sextractor.BLENDED):
            print "This star is BLENDED"
        star = catalog_f.readline()
    catalog_f.close()

-----------------------------------------------------------------


"""

# ======================================================================

from six.moves import builtins as __builtin__


# ======================================================================

__version__ = "0.1.5 (2005-02-14)"

# ======================================================================

# -- FLAGS meaning

NEIGHBOURS = 1
BLENDED = 2
SATURATED = 4
TRUNCATED = 8
CORRUPTED_APER = 16
CORRUPTED_ISO = 32
OVERFLOW_DEBLEND = 64
OVERFLOW_EXTRACT = 128


class WrongSExtractorfileException(Exception):
    pass


class SExtractorfile:

    """
    A class to manipulate SExtractor ASCII catalogs.

    For the moment only reading ('r' mode) is supported.

    """

    _SE_keys = \
        {"NUMBER": {"comment": "Running object number",
                    "infunc": int,
                    "format": "%10d",
                    "unit": ""},

         "FLAGS": {"comment": "Extraction flags",
                   "infunc": int,
                   "format": "%3d",
                   "unit": ""},

         "FLUX_ISO": {"comment": "Isophotal flux",
                      "infunc": float,
                      "format": "%12g",
                      "unit": "count"},

         "FLUXERR_ISO": {"comment": "RMS error for isophotal flux",
                         "infunc": float,
                         "format": "%12g",
                         "unit": "count"},

         "MAG_ISO": {"comment": "Isophotal magnitude",
                     "infunc": float,
                     "format": "%8.4f",
                     "unit": "mag"},

         "MAGERR_ISO": {"comment":
                        "RMS error for isophotal magnitude",
                        "infunc": float,
                        "format": "%8.4f",
                        "unit": "mag"},

         "FLUX_ISOCOR": {"comment": "Corrected isophotal flux",
                         "infunc": float,
                         "format": "%12g",
                         "unit": "count"},

         "FLUXERR_ISOCOR": {"comment":
                            "RMS error for corrected isophotal flux",
                            "infunc": float,
                            "format": "%12g",
                            "unit": "count"},

         "MAG_ISOCOR": {"comment": "Corrected isophotal magnitude",
                        "infunc": float,
                        "format": "%8.4f",
                        "unit": "mag"},

         "MAGERR_ISOCOR": {"comment":
                           "RMS error for corrected isophotal magnitude",
                           "infunc": float,
                           "format": "%8.4f",
                           "unit": "mag"},

         "FLUX_AUTO": {"comment":
                       "Flux within a Kron-like elliptical aperture",
                       "infunc": float,
                       "format": "%12g",
                       "unit": "count"},

         "FLUXERR_AUTO": {"comment": "RMS error for AUTO flux",
                          "infunc": float,
                          "format": "%12g",
                          "unit": "count"},

         "MAG_AUTO": {"comment":
                      "Kron-like elliptical aperture magnitude",
                      "infunc": float,
                      "format": "%8.4f",
                      "unit": "mag"},

         "MAGERR_AUTO": {"comment": "RMS error for AUTO magnitude",
                         "infunc": float,
                         "format": "%8.4f",
                         "unit": "mag"},

         "FLUX_BEST": {"comment":
                       "Best of FLUX_AUTO and FLUX_ISOCOR",
                       "infunc": float,
                       "format": "%12g",
                       "unit": "count"},

         "FLUXERR_BEST": {"comment": "RMS error for BEST flux",
                          "infunc": float,
                          "format": "%12g",
                          "unit": "count"},

         "MAG_BEST": {"comment": "Best of MAG_AUTO and MAG_ISOCOR",
                      "infunc": float,
                      "format": "%8.4f",
                      "unit": "mag"},

         "MAGERR_BEST": {"comment": "RMS error for MAG_BEST",
                         "infunc": float,
                         "format": "%8.4f",
                         "unit": "mag"},

         "KRON_RADIUS": {"comment":
                         "Kron apertures in units of A or B",
                         "infunc": float,
                         "format": "%5.2f",
                         "unit": ""},
         "BACKGROUND": {"comment": "Background at centroid position",
                        "infunc": float,
                        "format": "%12g",
                        "unit": "count"},

         "THRESHOLD": {"comment":
                       "Detection threshold above background",
                       "infunc": float,
                       "format": "%12g",
                       "unit": "count"},

         "MU_THRESHOLD": {"comment":
                          "Detection threshold above background",
                          "infunc": float,
                          "format": "%8.4f",
                          "unit": "mag * arcsec**(-2)"},

         "FLUX_MAX": {"comment": "Peak flux above background",
                      "infunc": float,
                      "format": "%12g",
                      "unit": "count"},

         "MU_MAX": {"comment":
                    "Peak surface brightness above background",
                    "infunc": float,
                    "format": "%8.4f",
                    "unit": "mag * arcsec**(-2)"},

         "ISOAREA_WORLD": {"comment":
                           "Isophotal area above Analysis threshold",
                           "infunc": float,
                           "format": "%12g",
                           "unit": "deg**2"},

         "XMIN_IMAGE": {"comment":
                        "Minimum x-coordinate among detected pixels",
                        "infunc": int,
                        "format": "%10d",
                        "unit": "pixel"},

         "YMIN_IMAGE": {"comment":
                        "Minimum y-coordinate among detected pixels",
                        "infunc": int,
                        "format": "%10d",
                        "unit": "pixel"},

         "XMAX_IMAGE": {"comment":
                        "Maximum x-coordinate among detected pixels",
                        "infunc": int,
                        "format": "%10d",
                        "unit": "pixel"},

         "YMAX_IMAGE": {"comment":
                        "Maximum y-coordinate among detected pixels",
                        "infunc": int,
                        "format": "%10d",
                        "unit": "pixel"},

         "X_IMAGE": {"comment": "Object position along x",
                     "infunc": float,
                     "format": "%10.3f",
                     "unit": "pixel"},

         "Y_IMAGE": {"comment": "Object position along y",
                     "infunc": float,
                     "format": "%10.3f",
                     "unit": "pixel"},

         "XWIN_IMAGE": {"comment": "Windowed position estimate along x",
                        "infunc": float,
                        "format": "%10.3f",
                        "unit": "pixel"},

         "YWIN_IMAGE": {"comment": "Windowed position estimate along y",
                        "infunc": float,
                        "format": "%10.3f",
                        "unit": "pixel"},
         "X_WORLD": {"comment":
                     "Barycenter position along world x axis",
                     "infunc": float,
                     "format": "%15e",
                     "unit": "deg"},

         "Y_WORLD": {"comment":
                     "Barycenter position along world y axis",
                     "infunc": float,
                     "format": "%15e",
                     "unit": "deg"},

         "ALPHA_SKY": {"comment":
                       "Right ascension of barycenter (native)",
                       "infunc": float,
                       "format": "%11.7f",
                       "unit": "deg"},

         "DELTA_SKY": {"comment":
                       "Declination of barycenter (native)",
                       "infunc": float,
                       "format": "%+11.7f",
                       "unit": "deg"},

         "ALPHA_J2000": {"comment":
                         "Right ascension of barycenter (J2000)",
                         "infunc": float,
                         "format": "%11.7f",
                         "unit": "deg"},

         "DELTA_J2000": {"comment":
                         "Declination of barycenter (J2000)",
                         "infunc": float,
                         "format": "%+11.7f",
                         "unit": "deg"},

         "ALPHA_B1950": {"comment":
                         "Right ascension of barycenter (B1950)",
                         "infunc": float,
                         "format": "%11.7f",
                         "unit": "deg"},

         "DELTA_B1950": {"comment":
                         "Declination of barycenter (B1950)",
                         "infunc": float,
                         "format": "%+11.7f",
                         "unit": "deg"},

         "X2_IMAGE": {"comment": "Variance along x",
                      "infunc": float,
                      "format": "%15e",
                      "unit": "pixel**2"},

         "Y2_IMAGE": {"comment": "Variance along y",
                      "infunc": float,
                      "format": "%15e",
                      "unit": "pixel**2"},

         "XY_IMAGE": {"comment": "Covariance between x and y",
                      "infunc": float,
                      "format": "%15e",
                      "unit": "pixel**2"},

         "CXX_IMAGE": {"comment": "Cxx object ellipse parameter",
                       "infunc": float,
                       "format": "%12e",
                       "unit": "pixel**(-2)"},

         "CYY_IMAGE": {"comment": "Cyy object ellipse parameter",
                       "infunc": float,
                       "format": "%12e",
                       "unit": "pixel**(-2)"},

         "CXY_IMAGE": {"comment": "Cxy object ellipse parameter",
                       "infunc": float,
                       "format": "%12e",
                       "unit": "pixel**(-2)"},

         "A_IMAGE": {"comment": "Profile RMS along major axis",
                     "infunc": float,
                     "format": "%9.3f",
                     "unit": "pixel"},

         "B_IMAGE": {"comment": "Profile RMS along minor axis",
                     "infunc": float,
                     "format": "%9.3f",
                     "unit": "pixel"},

         "THETA_IMAGE": {"comment": "Position angle (CCW/x)",
                         "infunc": float,
                         "format": "%5.1f",
                         "unit": "deg"},

         "ELONGATION": {"comment": "A_IMAGE/B_IMAGE",
                        "infunc": float,
                        "format": "%8.3f",
                        "unit": ""},

         "ELLIPTICITY": {"comment": "1 - B_IMAGE/A_IMAGE",
                         "infunc": float,
                         "format": "%8.3f",
                         "unit": ""},

         "ERRX2_IMAGE": {"comment": "Variance of position along x",
                         "infunc": float,
                         "format": "%15e",
                         "unit": "pixel**2"},

         "ERRY2_IMAGE": {"comment": "Variance of position along y",
                         "infunc": float,
                         "format": "%15e",
                         "unit": "pixel**2"},

         "ERRXY_IMAGE": {"comment":
                         "Covariance of position between x and y",
                         "infunc": float,
                         "format": "%15e",
                         "unit": "pixel**2"},

         "ERRCXX_IMAGE": {"comment": "Cxx error ellipse parameter",
                          "infunc": float,
                          "format": "%12g",
                          "unit": "pixel**(-2)"},

         "ERRCYY_IMAGE": {"comment": "Cyy error ellipse parameter",
                          "infunc": float,
                          "format": "%12g",
                          "unit": "pixel**(-2)"},

         "ERRCXY_IMAGE": {"comment": "Cxy error ellipse parameter",
                          "infunc": float,
                          "format": "%12g",
                          "unit": "pixel**(-2)"},

         "ERRA_IMAGE": {"comment":
                        "RMS position error along major axis",
                        "infunc": float,
                        "format": "%8.4f",
                        "unit": "pixel"},

         "ERRB_IMAGE": {"comment":
                        "RMS position error along minor axis",
                        "infunc": float,
                        "format": "%8.4f",
                        "unit": "pixel"},

         "ERRTHETA_IMAGE": {"comment":
                            "Error ellipse position angle (CCW/x)",
                            "infunc": float,
                            "format": "%5.1f",
                            "unit": "deg"},

         "FWHM_IMAGE": {"comment": "FWHM assuming a gaussian core",
                        "infunc": float,
                        "format": "%8.2f",
                        "unit": "pixel"},

         "X2_WORLD": {"comment": "Variance along X-WORLD (alpha)",
                      "infunc": float,
                      "format": "%15e",
                      "unit": "deg**2"},

         "Y2_WORLD": {"comment": "Variance along Y-WORLD (delta)",
                      "infunc": float,
                      "format": "%15e",
                      "unit": "deg**2"},

         "XY_WORLD": {"comment":
                      "Covariance between X-WORLD and Y-WORLD",
                      "infunc": float,
                      "format": "%15e",
                      "unit": "deg**2"},

         "CXX_WORLD": {"comment":
                       "Cxx object ellipse parameter (WORLD units)",
                       "infunc": float,
                       "format": "%12e",
                       "unit": "deg**(-2)"},

         "CYY_WORLD": {"comment":
                       "Cyy object ellipse parameter (WORLD units)",
                       "infunc": float,
                       "format": "%12e",
                       "unit": "deg**(-2)"},

         "CXY_WORLD": {"comment":
                       "Cxy object ellipse parameter (WORLD units)",
                       "infunc": float,
                       "format": "%12e",
                       "unit": "deg**(-2)"},

         "A_WORLD": {"comment":
                     "Profile RMS along major axis (world units)",
                     "infunc": float,
                     "format": "%12g",
                     "unit": "deg"},

         "B_WORLD": {"comment":
                     "Profile RMS along minor axis (world units)",
                     "infunc": float,
                     "format": "%12g",
                     "unit": "deg"},

         "THETA_WORLD": {"comment": "Position angle (CCW/world-x)",
                         "infunc": float,
                         "format": "%5.1f",
                         "unit": "deg"},

         "THETA_SKY": {"comment":
                       "Position angle (east of north) (native)",
                       "infunc": float,
                       "format": "%+6.2f",
                       "unit": "deg"},

         "THETA_J2000": {"comment":
                         "Position angle (east of north) (J2000)",
                         "infunc": float,
                         "format": "%+6.2f",
                         "unit": "deg"},

         "THETA_B1950": {"comment":
                         "Position angle (east of north) (B1950)",
                         "infunc": float,
                         "format": "%+6.2f",
                         "unit": "deg"},

         "ERRX2_WORLD": {"comment":
                         "Variance of position along X-WORLD (alpha)",
                         "infunc": float,
                         "format": "%15e",
                         "unit": "deg**2"},

         "ERRY2_WORLD": {"comment":
                         "Variance of position along Y-WORLD (delta)",
                         "infunc": float,
                         "format": "%15e",
                         "unit": "deg**2"},

         "ERRXY_WORLD": {"comment":
                         "Covariance of position X-WORLD/Y-WORLD",
                         "infunc": float,
                         "format": "%15e",
                         "unit": "deg**2"},

         "ERRCXX_WORLD": {"comment":
                          "Cxx error ellipse parameter (WORLD units)",
                          "infunc": float,
                          "format": "%12g",
                          "unit": "deg**(-2)"},

         "ERRCYY_WORLD": {"comment":
                          "Cyy error ellipse parameter (WORLD units)",
                          "infunc": float,
                          "format": "%12g",
                          "unit": "deg**(-2)"},

         "ERRCXY_WORLD": {"comment":
                          "Cxy error ellipse parameter (WORLD units)",
                          "infunc": float,
                          "format": "%12g",
                          "unit": "deg**(-2)"},

         "ERRA_WORLD": {"comment":
                        "World RMS position error along major axis",
                        "infunc": float,
                        "format": "%12g",
                        "unit": "pixel"},

         "ERRB_WORLD": {"comment":
                        "World RMS position error along minor axis",
                        "infunc": float,
                        "format": "%12g",
                        "unit": "pixel"},

         "ERRTHETA_WORLD": {"comment":
                            "Error ellipse pos. angle (CCW/world-x)",
                            "infunc": float,
                            "format": "%5.1f",
                            "unit": "deg"},

         "ERRTHETA_SKY": {"comment":
                          "Native error ellipse pos." +
                          "angle (east of north)",
                          "infunc": float,
                          "format": "%5.1f",
                          "unit": "deg"},

         "ERRTHETA_J2000": {"comment":
                            "J2000 error ellipse pos." +
                            "angle (east of north)",
                            "infunc": float,
                            "format": "%5.1f",
                            "unit": "deg"},

         "ERRTHETA_B1950": {"comment":
                            "B1950 error ellipse pos." +
                            "angle (east of north)",
                            "infunc": float,
                            "format": "%5.1f",
                            "unit": "deg"},

         "FWHM_WORLD": {"comment": "FWHM assuming a gaussian core",
                        "infunc": float,
                        "format": "%12g",
                        "unit": "deg"},

         "CLASS_STAR": {"comment": "S/G classifier output",
                        "infunc": float,
                        "format": "%5.2f",
                        "unit": ""}
         }

    def __init__(self, name, mode='r'):
        self.name = name
        self.mode = mode
        self.closed = True

        self._file = None
        self._keys = list()
        self._keys_positions = {}
        self._output = None
        self._firstline = True

        if self.mode != 'r':
            raise ValueError(
                'only read-only access is now implemented.'
                )

        self._file = __builtin__.open(self.name, self.mode)
        self.closed = False

        # Reading header

        self._line = self._file.readline()
        if not(self._line):
            raise WrongSExtractorfileException(
                'not a SExtractor text catalog (empty file)'
                )

        while (self._line):
            __ll = (self._line).replace('\n', '')
            if __ll[0] == '#':   # Still in header
                columns = __ll.split()
                if len(columns) < 3:
                    raise WrongSExtractorfileException(
                        'not a SExtractor text catalog (invalid header)'
                        )
                name = columns[2]
                if name not in SExtractorfile._SE_keys:
                    raise WrongSExtractorfileException(
                        'not a SExtractor text catalog (unknown keyword %s)'
                        % name
                        )
                self._keys_positions[name] = int(columns[1]) - 1
                self._keys.append(name)
            else:
                break
            self._line = self._file.readline()

        if not(self._keys):
            raise WrongSExtractorfileException(
                'not a SExtractor text catalog (empty header)'
                )

        self._outdict = dict([(k, None) for k in self._keys])
        self._firstline = True

    def __del__(self):
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        rr = self.readline()
        if not(rr):
            raise StopIteration
        return rr

    def __bool__(self):
        return self._file

    def keys(self):
        "Return the list of available parameters."
        return self._keys

    def getcolumns(self):
        "Return the list of available parameters."
        return list(self.keys())

    def readline(self):
        """
        Read and analyse the next line of the SExtractor catalog
        and return a dictionary {'param1': value, 'param2': value, ...}.
        """
        if not(self._firstline):
            self._line = self._file.readline()

        self._firstline = False

        if not(self._line):
            return None
        __ll = (self._line).replace('\n', '')
        __values = __ll.split()
        self._outdict.update(dict(zip(self._keys, __values)))
        for i in self._keys:
            self._outdict[i] = (
                SExtractorfile._SE_keys[i]["infunc"](self._outdict[i]))

        return self._outdict.copy()

    def read(self):
        """
        Read the file until EOF and return a list of dictionaries.
        """
        __result = []

        __ll = self.readline()
        while __ll:
            __result.append(__ll)
            __ll = self.readline()

        return list(__result)

    def readlines(self):
        return self.read()

    def close(self):
        """
        Close the SExtractor file.
        """
        if self._file:
            if not(self._file.closed):
                self._file.close()
        self.closed = True


# ======================================================================

def open(name, mode='r'):
    """
    Factory function.
    Open a SExtractor file and return a SExtractor file object.
    """
    return SExtractorfile(name, mode)

# ======================================================================



#
# Copyright 2013-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Recipe to detect slits in the AIV mask"""

from __future__ import division


import numpy
import six
from numina.core import Result, Parameter
from numina.exceptions import RecipeError
import numina.types.array as tarray
from numina.processing.combine import basic_processing_with_combination
from scipy import ndimage
from scipy.ndimage.filters import median_filter
from skimage.feature import canny

import emirdrp.datamodel as datamodel
from emirdrp.core.recipe import EmirRecipe
import emirdrp.requirements as reqs
import emirdrp.products as prods
from .common import normalize_raw, char_slit


class TestSlitDetectionRecipe(EmirRecipe):

    # Recipe Requirements
    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_sky = reqs.MasterSkyRequirement()

    median_filter_size = Parameter(5, 'Size of the median box')
    canny_sigma = Parameter(3.0, 'Sigma for the canny algorithm')
    canny_high_threshold = Parameter(0.04, 'High threshold for the Canny algorithm')
    canny_low_threshold = Parameter(0.01, 'High threshold for the Canny algorithm')

    # Recipe Results
    frame = Result(prods.ProcessedImage)
    slitstable = Result(tarray.ArrayType)
    DTU = Result(tarray.ArrayType)
    ROTANG = Result(float)
    DETPA = Result(float)
    DTUPA = Result(float)

    def run(self, rinput):
        self.logger.info('starting slit processing')

        self.logger.info('basic image reduction')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow=flow)
        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        try:
            rotang = hdr['ROTANG']
            detpa = hdr['DETPA']
            dtupa = hdr['DTUPA']
            dtub, dtur = datamodel.get_dtur_from_img(hdulist)

        except KeyError as error:
            self.logger.error(error)
            raise RecipeError(error)

        self.logger.debug('finding slits')


        # Filter values below 0.0
        self.logger.debug('Filter values below 0')
        data1 = hdulist[0].data[:]

        data1[data1 < 0.0] = 0.0
        # First, prefilter with median
        median_filter_size = rinput.median_filter_size
        canny_sigma = rinput.canny_sigma

        self.logger.debug('Median filter with box %d', median_filter_size)
        data2 = median_filter(data1, size=median_filter_size)

        # Grey level image
        img_grey = normalize_raw(data2)

        # Find edges with Canny
        self.logger.debug('Find edges, Canny sigma %f', canny_sigma)
        # These thresholds corespond roughly with
        # value x (2**16 - 1)
        high_threshold = rinput.canny_high_threshold
        low_threshold = rinput.canny_low_threshold
        self.logger.debug('Find edges, Canny high threshold %f', high_threshold)
        self.logger.debug('Find edges, Canny low threshold %f', low_threshold)
        edges = canny(img_grey, sigma=canny_sigma,
                      high_threshold=high_threshold,
                      low_threshold=low_threshold)
        
        # Fill edges
        self.logger.debug('Fill holes')
        # I do a dilation and erosion to fill
        # possible holes in 'edges'
        fill = ndimage.binary_dilation(edges)
        fill2 = ndimage.binary_fill_holes(fill)
        fill_slits = ndimage.binary_erosion(fill2)

        self.logger.debug('Label objects')
        label_objects, nb_labels = ndimage.label(fill_slits)
        self.logger.debug('%d objects found', nb_labels)
        ids = list(six.moves.range(1, nb_labels + 1))

        self.logger.debug('Find regions and centers')
        regions = ndimage.find_objects(label_objects)
        centers = ndimage.center_of_mass(data2, labels=label_objects,
                                         index=ids
                                         )

        table = char_slit(data2, regions,
                          slit_size_ratio=-1.0
                          )

        result = self.create_result(frame=hdulist,
                                    slitstable=table,
                                    DTU=dtub,
                                    ROTANG=rotang,
                                    DETPA=detpa,
                                    DTUPA=dtupa
                                    )

        return result


class TestSlitMaskDetectionRecipe(EmirRecipe):

    # Recipe Requirements
    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_sky = reqs.MasterSkyRequirement()

    median_filter_size = Parameter(5, 'Size of the median box')
    canny_sigma = Parameter(3.0, 'Sigma for the Canny algorithm')
    canny_high_threshold = Parameter(0.04, 'High threshold for the Canny algorithm')
    canny_low_threshold = Parameter(0.01, 'High threshold for the Canny algorithm')
    obj_min_size = Parameter(200, 'Minimum size of the slit')
    obj_max_size = Parameter(3000, 'Maximum size of the slit')
    slit_size_ratio = Parameter(4.0, 'Minimum ratio between height and width for slits')

    # Recipe Results
    frame = Result(prods.DataFrameType)
    slitstable = Result(tarray.ArrayType)
    DTU = Result(tarray.ArrayType)
    ROTANG = Result(float)
    DETPA = Result(float)
    DTUPA = Result(float)

    def run(self, rinput):
        self.logger.info('starting slit processing')

        self.logger.info('basic image reduction')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow=flow)
        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        try:
            rotang = hdr['ROTANG']
            detpa = hdr['DETPA']
            dtupa = hdr['DTUPA']
            dtub, dtur = datamodel.get_dtur_from_img(hdulist)

        except KeyError as error:
            self.logger.error(error)
            raise RecipeError(error)

        self.logger.debug('finding slits')

        # First, prefilter with median
        median_filter_size = rinput.median_filter_size
        canny_sigma = rinput.canny_sigma
        obj_min_size = rinput.obj_min_size
        obj_max_size = rinput.obj_max_size

        data1 = hdulist[0].data
        self.logger.debug('Median filter with box %d', median_filter_size)
        data2 = median_filter(data1, size=median_filter_size)

        # Grey level image
        img_grey = normalize_raw(data2)

        # Find edges with Canny
        self.logger.debug('Find edges with Canny, sigma %f', canny_sigma)
        # These thresholds corespond roughly with
        # value x (2**16 - 1)
        high_threshold = rinput.canny_high_threshold
        low_threshold = rinput.canny_low_threshold
        self.logger.debug('Find edges, Canny high threshold %f', high_threshold)
        self.logger.debug('Find edges, Canny low threshold %f', low_threshold)
        edges = canny(img_grey, sigma=canny_sigma,
                      high_threshold=high_threshold,
                      low_threshold=low_threshold)
        # Fill edges
        self.logger.debug('Fill holes')
        fill_slits =  ndimage.binary_fill_holes(edges)

        self.logger.debug('Label objects')
        label_objects, nb_labels = ndimage.label(fill_slits)
        self.logger.debug('%d objects found', nb_labels)
        # Filter on the area of the labeled region
        # Perhaps we could ignore this filtering and
        # do it later?
        self.logger.debug('Filter objects by size')
        # Sizes of regions
        sizes = numpy.bincount(label_objects.ravel())

        self.logger.debug('Min size is %d', obj_min_size)
        self.logger.debug('Max size is %d', obj_max_size)

        mask_sizes = (sizes > obj_min_size) & (sizes < obj_max_size)

        # Filter out regions
        nids, = numpy.where(mask_sizes)

        mm = numpy.in1d(label_objects, nids)
        mm.shape = label_objects.shape

        fill_slits_clean = numpy.where(mm, 1, 0)
        #plt.imshow(fill_slits_clean)

        # and relabel
        self.logger.debug('Label filtered objects')
        relabel_objects, nb_labels = ndimage.label(fill_slits_clean)
        self.logger.debug('%d objects found after filtering', nb_labels)
        ids = list(six.moves.range(1, nb_labels + 1))

        self.logger.debug('Find regions and centers')
        regions = ndimage.find_objects(relabel_objects)
        centers = ndimage.center_of_mass(data2, labels=relabel_objects,
                                         index=ids
                                         )

        table = char_slit(data2, regions,
                          slit_size_ratio=rinput.slit_size_ratio
                          )

        result = self.create_result(frame=hdulist, slitstable=table,
                                    DTU=dtub,
                                    ROTANG=rotang,
                                    DETPA=detpa,
                                    DTUPA=dtupa
                                    )

        return result




#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Auxiliary Recipes for EMIR"""

import logging

from numina.array.combine import median
from numina.core import DataFrame
from numina.core import Result
from numina.exceptions import RecipeError
from numina.util.context import manage_fits

from emirdrp.core import EMIR_BIAS_MODES
from emirdrp.core.recipe import EmirRecipe
from emirdrp.processing.info import gather_info_frames
import emirdrp.products as prods
import emirdrp.requirements as reqs

from numina.processing.combine import basic_processing_with_combination
from emirdrp.processing.combine import basic_processing_with_segmentation
from emirdrp.processing.combine import combine_images, scale_with_median


_logger = logging.getLogger('numina.recipes.emir')


class BiasRecipe(EmirRecipe):
    """
    Recipe to process data taken in Bias image Mode.

    Bias images only appear in Simple Readout mode.

    **Observing modes:**

     * Bias Image (3.1)

    **Inputs:**

    **Outputs:**

     * A combined bias frame, with variance extension.
     * Statistics of the final image per channel (mean, median, variance)

    **Procedure:**

    The list of images can be readly processed by
    combining them with a median algorithm.

    """
    
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    obresult = reqs.ObservationResultRequirement()

    biasframe = Result(prods.MasterBias)

    def run(self, rinput):
        _logger.info('starting bias reduction')

        iinfo = gather_info_frames(rinput.obresult.frames)

        if iinfo:
            mode = iinfo[0]['readmode']
            if mode.lower() not in EMIR_BIAS_MODES:
                msg = 'readmode %s, is not a bias mode' % mode
                _logger.error(msg)
                raise RecipeError(msg)

        flow = lambda x: x
        hdulist = basic_processing_with_combination(rinput, flow,
                                                    method=median,
                                                    errors=False)

        pdata = hdulist[0].data

        # update hdu header with
        # reduction keywords
        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        hdr['CCDMEAN'] = pdata.mean()

        _logger.info('bias reduction ended')

        result = self.create_result(biasframe=DataFrame(hdulist))
        return result


class DarkRecipe(EmirRecipe):
    """Recipe to process data taken in Dark current image Mode.

    Recipe to process dark images. The dark images will be combined
    using the median.
    They do have to be of the same exposure time t.

    **Observing mode:**

     * Dark current Image (3.2)

    **Inputs:**

    **Outputs:**

     * A combined dark frame, with variance extension.
    """

    master_bpm = reqs.MasterBadPixelMaskRequirement()
    obresult = reqs.ObservationResultRequirement()
    master_bias = reqs.MasterBiasRequirement()

    darkframe = Result(prods.MasterDark)

    def run(self, rinput):

        _logger.info('starting dark reduction')

        flow = self.init_filters(rinput)

        iinfo = gather_info_frames(rinput.obresult.frames)
        ref_exptime = 0.0
        for el in iinfo[1:]:
            if abs(el['texp'] - ref_exptime) > 1e-4:
                _logger.error('image with wrong exposure time')
                raise RecipeError('image with wrong exposure time')

        hdulist = basic_processing_with_combination(rinput, flow,
                                                    method=median,
                                                    errors=True)

        pdata = hdulist[0].data

        # update hdu header with
        # reduction keywords

        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        hdr['CCDMEAN'] = pdata.mean()

        _logger.info('dark reduction ended')
        result = self.create_result(darkframe=hdulist)
        return result


class IntensityFlatRecipe(EmirRecipe):
    """Recipe to process data taken in intensity flat-field mode.

    Recipe to process intensity flat-fields. The flat-on and
    flat-off images are combined (method?) separately and the subtracted
    to obtain a thermal subtracted flat-field.

    **Observing modes:**

     * Intensity Flat-Field

    **Inputs:**

      * A master dark frame
      * Non linearity
      * A model of the detector.

    **Outputs:**

     * TBD

    **Procedure:**

     * A combined thermal subtracted flat field, normalized to median 1,
       with with variance extension and quality flag.

    """

    master_bpm = reqs.MasterBadPixelMaskRequirement()
    obresult = reqs.ObservationResultRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()

    flatframe = Result(prods.MasterIntensityFlat)

    def run(self, rinput):
        _logger.info('starting flat reduction')

        errors = True

        flow = self.init_filters(rinput)
        hdulist = basic_processing_with_combination(rinput, flow,
                                                    method=median,
                                                    errors=errors)

        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        mm = hdulist[0].data.mean()
        hdr['CCDMEAN'] = mm

        hdulist[0].data /= mm
        if errors:
            hdulist['variance'].data /= (mm*mm)

        result = self.create_result(flatframe=hdulist)

        return result


class IntensityFlatRecipe2(EmirRecipe):
    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()

    master_flatframe = Result(prods.MasterIntensityFlat)

    def run(self, rinput):
        from numina.array import combine

        _logger.info('starting flat reduction')

        frames = rinput.obresult.frames

        with manage_fits(frames) as list_of:
            scaled_mean = scale_with_median(combine.mean)
            c_img = combine_images(list_of, method=scaled_mean, errors=False)

        self.save_intermediate_img(c_img, 'p0.fits')

        flow = self.init_filters(rinput)

        processed_img = flow(c_img)

        self.save_intermediate_img(processed_img, 'p1.fits')

        hdr = processed_img[0].header
        self.set_base_headers(hdr)

        import scipy.ndimage.filters

        _logger.info('median filter')
        data_smooth = scipy.ndimage.filters.median_filter(processed_img[0].data, size=11)

        self.save_intermediate_array(data_smooth, 'smooth.fits')

        mm = processed_img[0].data.mean()
        hdr['CCDMEAN'] = mm

        processed_img[0].data /= data_smooth

        self.save_intermediate_img(processed_img, 'p2.fits')

        result = self.create_result(master_flatframe=processed_img)

        return result


class SimpleSkyRecipe(EmirRecipe):
    """Recipe to process data taken in intensity flat-field mode.

    """

    master_bpm = reqs.MasterBadPixelMaskRequirement()
    obresult = reqs.ObservationResultRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()

    skyframe = Result(prods.MasterSky)


    def run(self, rinput):
        _logger.info('starting sky reduction')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow,
                                                    method=median,
                                                    errors=True)

        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        result = self.create_result(skyframe=hdulist)

        return result


class DitherSkyRecipe(EmirRecipe):
    """Recipe to process data taken in dither sky mode.

    """

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()

    skyframe = Result(prods.MasterSky)


    def run(self, rinput):
        _logger.debug('instrument %s, mode %s', rinput.obresult.instrument,
                      rinput.obresult.mode
                      )
        _logger.info('starting sky reduction with dither')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_segmentation(rinput, flow,
                                                    method=median,
                                                    errors=True)

        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        _logger.info('end sky reduction with dither')

        result = self.create_result(skyframe=hdulist)

        return result


class SpectralFlatRecipe(EmirRecipe):
    """Recipe to process data taken in intensity flat-field mode."""

    master_bpm = reqs.MasterBadPixelMaskRequirement()
    obresult = reqs.ObservationResultRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    
    flatframe = Result(prods.MasterSpectralFlat)

    def run(self, rinput):
        return self.create_result(flatframe=prods.MasterSpectralFlat())


class SlitTransmissionRecipe(EmirRecipe):
    """Recipe to calibrate the slit transmission.

    **Observing modes:**

        * Slit transmission calibration (4.4)

    **Inputs:**

        * A list of uniformly illuminated images of MSM

    **Outputs:**

     * A list of slit transmission functions

    **Procedure:**

     * TBD

    """

    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()

    slit = Result(prods.SlitTransmissionCalibration)

    def run(self, rinput):
        return self.create_result(slit=prods.SlitTransmissionCalibration())



class WavelengthCalibrationRecipe(EmirRecipe):
    """Recipe to calibrate the spectral response.

    **Observing modes:**

        * Wavelength calibration (4.5)

    **Inputs:**

     * List of line positions
     * Calibrations up to spectral flatfielding

    **Outputs:**

     * Wavelength calibration structure

    **Procedure:**

     * TBD
    """

    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_spectral_ff = reqs.MasterSpectralFlatFieldRequirement()

    cal = Result(prods.WavelengthCalibration)

    def run(self, rinput):
        return self.create_result(cal=prods.WavelengthCalibration())



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
import numpy as np
from numpy.polynomial import Polynomial
import pkgutil
from six import StringIO
import sys

from emirdrp.processing.wavecal.set_wv_parameters import set_wv_parameters
from emirdrp.products import RectWaveCoeff

from numina.array.distortion import fmap

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NBARS

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES


def save_ds9(output, filename):
    """Save ds9 region output info filename.

    Parameters
    ----------
    output : str
        String containing the full output to be exported as a ds9 region
        file.
    filename : str
        Output file name.

    """

    ds9_file = open(filename, 'wt')
    ds9_file.write(output)
    ds9_file.close()


def save_four_ds9(rectwv_coeff, debugplot=0):
    """Save the 4 possible ds9 region files.

    Parameters
    ----------
    rectwv_coeff : RectWaveCoeff instance
        Rectification and wavelength calibration coefficients for the
        particular CSU configuration.
    debugplot : int
            Debugging level for messages and plots. For details see
            'numina.array.display.pause_debugplot.py'.

    """

    for limits, rectified, suffix in zip(
        ['frontiers', 'frontiers', 'boundaries', 'boundaries'],
        [False, True, False, True],
        ['rawimage', 'rectified', 'rawimage', 'rectified']
    ):
        output = rectwv_coeff_to_ds9(rectwv_coeff=rectwv_coeff,
                                     limits=limits,
                                     rectified=rectified)
        filename = 'ds9_' + limits + '_' + suffix + '.reg'
        if abs(debugplot) >= 10:
            print('>>> Saving: ', filename)
        save_ds9(output, filename)


def rectwv_coeff_to_ds9(rectwv_coeff,
                        limits=None,
                        rectified=False,
                        numpix=100):
    """Generate ds9 region output with requested slitlet limits

    Parameters
    ----------
    rectwv_coeff : RectWaveCoeff instance
        Rectification and wavelength calibration coefficients for the
        particular CSU configuration.
    limits : str
        Region to be saved: the only two possibilities are 'boundaries'
        and 'frontiers'.
    rectified : bool
        If True, the regions correspond to the rectified image.
    numpix : int
        Number of points in which the X-range interval is subdivided
        in order to save each boundary as a connected set of line
        segments. This option is only relevant when rectified=False.

    Returns
    -------
    output : str
        String containing the full output to be exported as a ds9 region
        file.
    """

    # protections
    if limits not in ['boundaries', 'frontiers']:
        raise ValueError('Unexpect limits=' + str(limits))

    # retrieve relevant wavelength calibration parameters
    grism_name = rectwv_coeff.tags['grism']
    filter_name = rectwv_coeff.tags['filter']
    wv_parameters = set_wv_parameters(filter_name, grism_name)
    naxis1_enlarged = wv_parameters['naxis1_enlarged']
    crpix1_enlarged = wv_parameters['crpix1_enlarged']
    crval1_enlarged = wv_parameters['crval1_enlarged']
    cdelt1_enlarged = wv_parameters['cdelt1_enlarged']

    ds9_output = '# Region file format: DS9 version 4.1\n' \
                 'global color=green dashlist=2 4 width=2 ' \
                 'font="helvetica 10 normal roman" select=1 ' \
                 'highlite=1 dash=1 fixed=0 edit=1 ' \
                 'move=1 delete=1 include=1 source=1\nphysical\n#\n'

    ds9_output += '#\n# uuid (rectwv_coeff): {0}\n'.format(rectwv_coeff.uuid)
    ds9_output += \
        '# grism...............: {0}\n'.format(rectwv_coeff.tags['grism'])
    ds9_output += \
        '# filter..............: {0}\n'.format(rectwv_coeff.tags['filter'])

    for islitlet in range(1, EMIR_NBARS + 1):
        if islitlet not in rectwv_coeff.missing_slitlets:
            dumdict = rectwv_coeff.contents[islitlet - 1]
            if islitlet % 2 == 0:
                if limits == 'frontiers':
                    colorbox = '#0000ff'  # '#ff77ff'
                else:
                    colorbox = '#ff00ff'  # '#ff77ff'
            else:
                if limits == 'frontiers':
                    colorbox = '#0000ff'  # '#4444ff'
                else:
                    colorbox = '#00ffff'  # '#4444ff'

            ds9_output += '#\n# islitlet...........: {0}\n'.format(islitlet)
            ds9_output += '# csu_bar_slit_center: {0}\n'.format(
                dumdict['csu_bar_slit_center']
            )
            if rectified:
                crpix1_linear = 1.0
                crval1_linear = dumdict['crval1_linear']
                cdelt1_linear = dumdict['cdelt1_linear']
                if limits == 'frontiers':
                    ydum_lower = dumdict['y0_frontier_lower_expected']
                    ydum_upper = dumdict['y0_frontier_upper_expected']
                else:
                    ydum_lower = dumdict['y0_reference_lower_expected']
                    ydum_upper = dumdict['y0_reference_upper_expected']
                wave_ini = crval1_linear + \
                           (0.5 - crpix1_linear) * cdelt1_linear
                xdum_ini = (wave_ini - crval1_enlarged) / cdelt1_enlarged
                xdum_ini += crpix1_enlarged
                wave_end = crval1_linear + \
                           (EMIR_NAXIS1 + 0.5 - crpix1_linear) * cdelt1_linear
                xdum_end = (wave_end - crval1_enlarged) / cdelt1_enlarged
                xdum_end += crpix1_enlarged
                for ydum in [ydum_lower, ydum_upper]:
                    ds9_output += \
                        'line {0} {1} {2} {3}'.format(
                            xdum_ini, ydum,
                            xdum_end, ydum
                        )
                    ds9_output += ' # color={0}\n'.format(colorbox)
                # slitlet label
                ydum_label = (ydum_lower + ydum_upper) / 2.0
                xdum_label = EMIR_NAXIS1 / 2 + 0.5
                wave_center = crval1_linear + \
                              (xdum_label - crpix1_linear) * cdelt1_linear
                xdum_label = (wave_center - crval1_enlarged) / cdelt1_enlarged
                xdum_label += crpix1_enlarged
                ds9_output += 'text {0} {1} {{{2}}} # color={3} ' \
                              'font="helvetica 10 bold ' \
                              'roman"\n'.format(xdum_label, ydum_label,
                                                islitlet, colorbox)
            else:
                if limits == 'frontiers':
                    pol_lower = Polynomial(
                        dumdict['frontier']['poly_coef_lower']
                    )
                    pol_upper = Polynomial(
                        dumdict['frontier']['poly_coef_upper']
                    )
                else:
                    pol_lower = Polynomial(
                        dumdict['spectrail']['poly_coef_lower']
                    )
                    pol_upper = Polynomial(
                        dumdict['spectrail']['poly_coef_upper']
                    )
                xdum = np.linspace(1, EMIR_NAXIS1, num=numpix)
                ydum = pol_lower(xdum)
                ydum -= float(rectwv_coeff.global_integer_offset_y_pix)
                for i in range(len(xdum) - 1):
                    ds9_output += \
                        'line {0} {1} {2} {3}'.format(
                            xdum[i], ydum[i],
                            xdum[i + 1], ydum[i + 1]
                        )
                    ds9_output += ' # color={0}\n'.format(colorbox)
                ydum = pol_upper(xdum)
                ydum -= float(rectwv_coeff.global_integer_offset_y_pix)
                for i in range(len(xdum) - 1):
                    ds9_output += \
                        'line {0} {1} {2} {3}'.format(
                            xdum[i], ydum[i],
                            xdum[i + 1], ydum[i + 1]
                        )
                    ds9_output += ' # color={0}\n'.format(colorbox)
                # slitlet label
                xdum_label = EMIR_NAXIS1 / 2 + 0.5
                ydum_lower = pol_lower(xdum_label)
                ydum_upper = pol_upper(xdum_label)
                ydum_label = (ydum_lower + ydum_upper) / 2.0
                ds9_output += 'text {0} {1} {{{2}}} # color={3} ' \
                              'font="helvetica 10 bold ' \
                              'roman"\n'.format(xdum_label, ydum_label,
                                                islitlet, colorbox)

    return ds9_output


def save_spectral_lines_ds9(rectwv_coeff, debugplot=0):
    """Save expected location of arc and OH airglow to ds9 region files.

    Parameters
    ----------
    rectwv_coeff : RectWaveCoeff instance
        Rectification and wavelength calibration coefficients for the
        particular CSU configuration.
    debugplot : int
            Debugging level for messages and plots. For details see
            'numina.array.display.pause_debugplot.py'.

    """

    for spectral_lines, rectified, suffix in zip(
        ['arc', 'arc', 'oh', 'oh'],
        [False, True, False, True],
        ['rawimage', 'rectified', 'rawimage', 'rectified']
    ):
        output = spectral_lines_to_ds9(rectwv_coeff=rectwv_coeff,
                                       spectral_lines=spectral_lines,
                                       rectified=rectified)
        filename = 'ds9_' + spectral_lines + '_' + suffix + '.reg'
        if abs(debugplot) >= 10:
            print('>>> Saving: ', filename)
        save_ds9(output, filename)


def spectral_lines_to_ds9(rectwv_coeff,
                    spectral_lines=None,
                    rectified=False):
    """Generate ds9 region output with requested spectral lines

    Parameters
    ----------
    rectwv_coeff : RectWaveCoeff instance
        Rectification and wavelength calibration coefficients for the
        particular CSU configuration.
    spectral_lines : str
        Spectral lines to be saved: the only two possibilities are 'arc'
        and 'oh'.
    rectified : bool
        If True, the regions correspond to the rectified image.

    Returns
    -------
    output : str
        String containing the full output to be exported as a ds9 region
        file.
    """

    # protections
    if spectral_lines not in ['arc', 'oh']:
        raise ValueError('Unexpected spectral lines=' + str(spectral_lines))

    if spectral_lines == 'arc':
        grism_name = rectwv_coeff.tags['grism']
        if grism_name == 'LR':
            catlines_file = 'lines_argon_neon_xenon_empirical_LR.dat'
        else:
            catlines_file = 'lines_argon_neon_xenon_empirical.dat'
        dumdata = pkgutil.get_data('emirdrp.instrument.configs',
                                   catlines_file)
        arc_lines_tmpfile = StringIO(dumdata.decode('utf8'))
        catlines = np.genfromtxt(arc_lines_tmpfile)
        # define wavelength and flux as separate arrays
        catlines_all_wave = catlines[:, 0]
        catlines_all_flux = catlines[:, 1]
    elif spectral_lines == 'oh':
        dumdata = pkgutil.get_data('emirdrp.instrument.configs',
                                   'Oliva_etal_2013.dat')
        oh_lines_tmpfile = StringIO(dumdata.decode('utf8'))
        catlines = np.genfromtxt(oh_lines_tmpfile)
        # define wavelength and flux as separate arrays
        catlines_all_wave = np.concatenate((catlines[:, 1],
                                            catlines[:, 0]))
        catlines_all_flux = np.concatenate((catlines[:, 2],
                                            catlines[:, 2]))
    else:
        raise ValueError('This should not happen!')

    # retrieve relevant wavelength calibration parameters
    grism_name = rectwv_coeff.tags['grism']
    filter_name = rectwv_coeff.tags['filter']
    wv_parameters = set_wv_parameters(filter_name, grism_name)
    naxis1_enlarged = wv_parameters['naxis1_enlarged']
    crpix1_enlarged = wv_parameters['crpix1_enlarged']
    crval1_enlarged = wv_parameters['crval1_enlarged']
    cdelt1_enlarged = wv_parameters['cdelt1_enlarged']

    ds9_output = '# Region file format: DS9 version 4.1\n' \
                 'global color=#00ffff dashlist=0 0 width=2 ' \
                 'font="helvetica 10 normal roman" select=1 ' \
                 'highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 ' \
                 'include=1 source=1\nphysical\n'

    ds9_output += '#\n# uuid (rectwv_coeff): {0}\n'.format(
        rectwv_coeff.uuid)
    ds9_output += \
        '# grism...............: {0}\n'.format(rectwv_coeff.tags['grism'])
    ds9_output += \
        '# filter..............: {0}\n'.format(rectwv_coeff.tags['filter'])

    global_integer_offset_x_pix = rectwv_coeff.global_integer_offset_x_pix
    global_integer_offset_y_pix = rectwv_coeff.global_integer_offset_y_pix

    for islitlet in range(1, EMIR_NBARS + 1):
        if islitlet not in rectwv_coeff.missing_slitlets:
            dumdict = rectwv_coeff.contents[islitlet - 1]
            if islitlet % 2 == 0:
                colorbox = '#ff00ff'  # '#ff77ff'
            else:
                colorbox = '#00ffff'  # '#4444ff'

            ds9_output += '#\n# islitlet...........: {0}\n'.format(
                islitlet)
            ds9_output += '# csu_bar_slit_center: {0}\n'.format(
                dumdict['csu_bar_slit_center']
            )
            crpix1_linear = 1.0
            crval1_linear = dumdict['crval1_linear']
            cdelt1_linear = dumdict['cdelt1_linear']
            wave_ini = crval1_linear + \
                       (0.5 - crpix1_linear) * cdelt1_linear
            wave_end = crval1_linear + \
                       (EMIR_NAXIS1 + 0.5 - crpix1_linear) * cdelt1_linear
            if rectified:
                ydum_lower = dumdict['y0_reference_lower_expected']
                ydum_upper = dumdict['y0_reference_upper_expected']
                # spectral lines
                for wave in catlines_all_wave:
                    if wave_ini <= wave <= wave_end:
                        xdum = (wave - crval1_enlarged) / cdelt1_enlarged
                        xdum += crpix1_enlarged
                        ds9_output += \
                            'line {0} {1} {2} {3}'.format(
                                xdum, ydum_lower,
                                xdum, ydum_upper
                            )
                        ds9_output += ' # color={0}\n'.format(colorbox)
                # slitlet label
                ydum_label = (ydum_lower + ydum_upper) / 2.0
                xdum_label = EMIR_NAXIS1 / 2 + 0.5
                wave_center = crval1_linear + \
                              (xdum_label - crpix1_linear) * cdelt1_linear
                xdum_label = (wave_center - crval1_enlarged) / cdelt1_enlarged
                xdum_label += crpix1_enlarged
                ds9_output += 'text {0} {1} {{{2}}} # color={3} ' \
                              'font="helvetica 10 bold ' \
                              'roman"\n'.format(xdum_label, ydum_label,
                                                islitlet, colorbox)
            else:
                bb_ns1_orig = dumdict['bb_ns1_orig']
                ttd_order = dumdict['ttd_order']
                aij = dumdict['ttd_aij']
                bij = dumdict['ttd_bij']
                min_row_rectified = float(dumdict['min_row_rectified'])
                max_row_rectified = float(dumdict['max_row_rectified'])
                mean_row_rectified = (min_row_rectified + max_row_rectified)/2
                wpoly_coeff = dumdict['wpoly_coeff']
                x0 = []
                y0 = []
                x1 = []
                y1 = []
                x2 = []
                y2 = []
                # spectral lines
                for wave in catlines_all_wave:
                    if wave_ini <= wave <= wave_end:
                        tmp_coeff = np.copy(wpoly_coeff)
                        tmp_coeff[0] -= wave
                        tmp_xroots = np.polynomial.Polynomial(
                            tmp_coeff).roots()
                        for dum in tmp_xroots:
                            if np.isreal(dum):
                                dum = dum.real
                                if 1 <= dum <= EMIR_NAXIS1:
                                    x0.append(dum)
                                    y0.append(mean_row_rectified)
                                    x1.append(dum)
                                    y1.append(min_row_rectified)
                                    x2.append(dum)
                                    y2.append(max_row_rectified)
                        pass
                xx0, yy0 = fmap(ttd_order, aij, bij, np.array(x0),
                                np.array(y0))
                xx0 -= global_integer_offset_x_pix
                yy0 += bb_ns1_orig
                yy0 -= global_integer_offset_y_pix
                xx1, yy1 = fmap(ttd_order, aij, bij, np.array(x1),
                                np.array(y1))
                xx1 -= global_integer_offset_x_pix
                yy1 += bb_ns1_orig
                yy1 -= global_integer_offset_y_pix
                xx2, yy2 = fmap(ttd_order, aij, bij, np.array(x2),
                                np.array(y2))
                xx2 -= global_integer_offset_x_pix
                yy2 += bb_ns1_orig
                yy2 -= global_integer_offset_y_pix
                for xx1_, xx2_, yy1_, yy2_ in zip(xx1, xx2, yy1, yy2):
                    ds9_output += \
                        'line {0} {1} {2} {3}'.format(
                            xx1_, yy1_, xx2_, yy2_
                        )
                    ds9_output += ' # color={0}\n'.format(colorbox)
                # slitlet label
                pol_lower = Polynomial(dumdict['spectrail']['poly_coef_lower'])
                pol_upper = Polynomial(dumdict['spectrail']['poly_coef_upper'])
                xdum_label = EMIR_NAXIS1 / 2 + 0.5
                ydum_lower = pol_lower(xdum_label)
                ydum_upper = pol_upper(xdum_label)
                ydum_label = (ydum_lower + ydum_upper) / 2.0
                ds9_output += 'text {0} {1} {{{2}}} # color={3} ' \
                              'font="helvetica 10 bold ' \
                              'roman"\n'.format(xdum_label, ydum_label,
                                                islitlet, colorbox)

    return ds9_output


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: generate ds9 region files associated '
                    'to a particular rectification and wavelength calibration'
    )

    # required arguments
    parser.add_argument("--rectwv_coeff", required=True,
                        help="Input JSON file with rectification and "
                             "wavelength calibration coefficients",
                        type=argparse.FileType('rt'))

    # optional arguments
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting & debugging options"
                             " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # generate RectWaveCoeff object
    rectwv_coeff = RectWaveCoeff._datatype_load(
        args.rectwv_coeff.name)

    save_four_ds9(rectwv_coeff=rectwv_coeff, debugplot=args.debugplot)
    save_spectral_lines_ds9(rectwv_coeff=rectwv_coeff,
                            debugplot=args.debugplot)


if __name__ == "__main__":

    main()



#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Routines shared by image mode recipes"""

import os
import logging
import shutil
import math

import six
import numpy
from astropy.io import fits
import astropy.wcs
from astropy.visualization import SqrtStretch
from astropy.visualization import PercentileInterval
from astropy.visualization.mpl_normalize import ImageNormalize
from scipy.spatial import KDTree as KDTree
import sep

import matplotlib as mpl
from matplotlib.patches import Ellipse
from matplotlib.collections import PatchCollection

from numina import __version__
from numina.core import DataFrame
from numina.array import fixpix2
from numina.frame import resize_fits, custom_region_to_str
from numina.array import combine_shape, correct_flatfield
from numina.array import subarray_match
from numina.array.combine import flatcombine, median, quantileclip
from numina.exceptions import RecipeError

from emirdrp.util.sextractor import SExtractor
from emirdrp.util.sextractor import open as sopen
from emirdrp.core.recipe import EmirRecipe
from emirdrp.products import SourcesCatalog
from emirdrp.instrument.channels import FULL
from emirdrp.processing.wcs import offsets_from_wcs

from .checks import check_photometry
from .naming import (name_redimensioned_frames, name_object_mask,
                     name_skybackground)
from .naming import name_skybackgroundmask, name_skysub_proc, name_skyflat
from .naming import name_skyflat_proc, name_segmask


_logger = logging.getLogger('numina.recipes.emir')


def intersection(a, b, scale=1):
    '''Intersection between two segments.'''
    try:
        a1, a2 = a
    except TypeError:
        a1 = a.start
        a2 = a.stop

    try:
        b1, b2 = b
    except TypeError:
        b1 = b.start
        b2 = b.stop

    if a2 <= b1:
        return None
    if a1 >= b2:
        return None

    # a2 > b1 and a1 < b2

    if a2 <= b2:
        if a1 <= b1:
            return slice(b1 * scale, a2 * scale)
        else:
            return slice(a1 * scale, a2 * scale)
    else:
        if a1 <= b1:
            return slice(b1 * scale, b2 * scale)
        else:
            return slice(a1 * scale, b2 * scale)


def clip_slices(r, region, scale=1):
    '''Intersect slices with a region.'''
    t = []
    for ch in r:
        a1 = intersection(ch[0], region[0], scale=scale)
        if a1 is None:
            continue
        a2 = intersection(ch[1], region[1], scale=scale)
        if a2 is None:
            continue

        t.append((a1, a2))

    return t


class DirectImageCommon(EmirRecipe):

    logger = _logger
    BASIC, PRERED, CHECKRED, FULLRED, COMPLETE = [0, 1, 2, 3, 4]
    __version__ = '1'

    def __init__(self, *args, **kwds):
        super(DirectImageCommon, self).__init__(version=__version__)
        # Required to delay the backend initialization (issue numina/#102, #49)
        import matplotlib.pyplot as plt

        self._figure = plt.figure(facecolor='white')
        self._figure.canvas.set_window_title('Recipe Plots')
        self._figure.canvas.draw()

    def process(self, ri,
                window=None, subpix=1,
                store_intermediate=True,
                target_is_sky=True, stop_after=PRERED):

        numpy.seterr(divide='raise')

        recipe_input = ri
        # FIXME: hardcoded instrument information
        keywords = {'airmass': 'AIRMASS',
                    'exposure': 'EXPTIME',
                    'imagetype': 'IMGTYP',
                    'juliandate': 'MJD-OBS',
                    'tstamp': 'TSTAMP'
                    }
        baseshape = [2048, 2048]
        channels = FULL

        if window is None:
            window = tuple((0, siz) for siz in baseshape)

        if store_intermediate:
            pass

        # States
        sf_data = None
        state = self.BASIC
        step = 0

        try:
            niteration = ri.iterations
        except KeyError:
            niteration = 1

        while True:
            if state == self.BASIC:
                _logger.info('Basic processing')

                # Basic processing
                basicflow = self.init_filters(recipe_input)

                for frame in ri.obresult.frames:
                    with frame.open() as hdulist:
                        hdulist = basicflow(hdulist)

                if stop_after == state:
                    break
                else:
                    state = self.PRERED
            elif state == self.PRERED:
                # Shape of the window
                windowshape = tuple((i[1] - i[0]) for i in window)
                _logger.debug('Shape of window is %s', windowshape)
                # Shape of the scaled window
                subpixshape = tuple((side * subpix) for side in windowshape)

                # Scaled window region
                scalewindow = tuple(
                    slice(*(subpix * i for i in p)) for p in window)
                # Window region
                window = tuple(slice(*p) for p in window)

                scaled_chan = clip_slices(channels, window, scale=subpix)

                # Reference pixel in the center of the frame
                refpix = numpy.divide(
                    numpy.array([baseshape], dtype='int'), 2).astype('float')

                # lists of targets and sky frames
                targetframes = []
                skyframes = []

                for frame in ri.obresult.frames:

                    # Getting some metadata from FITS header
                    hdr = fits.getheader(frame.label)
                    try:
                        frame.exposure = hdr[str(keywords['exposure'])]
                        # frame.baseshape = get_image_shape(hdr)
                        frame.airmass = hdr[str(keywords['airmass'])]
                        frame.mjd = hdr[str(keywords['tstamp'])]
                    except KeyError as e:
                        raise KeyError("%s in frame %s" %
                                       (str(e), frame.label))

                    frame.baselabel = os.path.splitext(frame.label)[0]
                    frame.mask = ri.master_bpm
                    # Insert pixel offsets between frames
                    frame.objmask_data = None
                    frame.valid_target = False
                    frame.valid_sky = False
                    frame.valid_region = scalewindow
                    # FIXME: hardcode itype for the moment
                    frame.itype = 'TARGET'
                    if frame.itype == 'TARGET':
                        frame.valid_target = True
                        targetframes.append(frame)
                        if target_is_sky:
                            frame.valid_sky = True
                            skyframes.append(frame)
                    if frame.itype == 'SKY':
                        frame.valid_sky = True
                        skyframes.append(frame)

#                labels = [frame.label for frame in targetframes]

                if ri.offsets is not None:
                    _logger.info('Using offsets from parameters')
                    base_ref = numpy.asarray(ri.offsets)
                    list_of_offsets= -(base_ref - base_ref[0])
                else:
                    _logger.info('Computing offsets from WCS information')
                    list_of_offsets = offsets_from_wcs(targetframes, refpix)

                # FIXME: Im using offsets in row/columns
                # the values are provided in XY so flip-lr
                list_of_offsets = numpy.fliplr(list_of_offsets)

                # Insert pixel offsets between frames
                for frame, off in zip(targetframes, list_of_offsets):

                    # Insert pixel offsets between frames
                    frame.pix_offset = off
                    frame.scaled_pix_offset = subpix * off

                    _logger.debug('Frame %s, offset=%s, scaled=%s',
                                  frame.label, off, subpix * off)

                _logger.info('Computing relative offsets')
                offsets = [(frame.scaled_pix_offset)
                           for frame in targetframes]
                offsets = numpy.round(offsets).astype('int')
                finalshape, offsetsp = combine_shape(subpixshape, offsets)
                _logger.info('Shape of resized array is %s', finalshape)

                # Resizing target frames
                self.resize(targetframes, subpixshape, offsetsp, finalshape,
                            window=window, scale=subpix)

                if not target_is_sky:
                    for frame in skyframes:
                        frame.resized_base = frame.label
                        frame.resized_mask = frame.mask

                # superflat
                _logger.info('Step %d, superflat correction (SF)', step)
                # Compute scale factors (median)
                self.update_scale_factors(ri.obresult.frames)

                # Create superflat
                superflat = self.compute_superflat(skyframes,
                                                   channels=scaled_chan,
                                                   step=step)

                # Apply superflat
                self.figure_init(subpixshape)
                self.apply_superflat(ri.obresult.frames, superflat)

                _logger.info('Simple sky correction')
                if target_is_sky:
                    # Each frame is the closest sky frame available

                    for frame in ri.obresult.frames:
                        self.compute_simple_sky_for_frame(frame, frame)
                else:
                    self.compute_simple_sky(targetframes, skyframes)

                # Combining the frames
                _logger.info("Step %d, Combining target frames", step)

                sf_data = self.combine_frames(
                    targetframes, extinction=ri.extinction)

                self.figures_after_combine(sf_data)

                _logger.info('Step %d, finished', step)

                if stop_after == state:
                    break
                else:
                    state = self.CHECKRED
            elif state == self.CHECKRED:

                seeing_fwhm = None

                # self.check_position(images_info, sf_data, seeing_fwhm)
                recompute = False
                if recompute:
                    _logger.info('Recentering is needed')
                    state = self.PRERED
                else:
                    _logger.info('Recentering is not needed')
                    _logger.info('Checking photometry')
                    check_photometry(targetframes, sf_data,
                                     seeing_fwhm, figure=self._figure)

                    if stop_after == state:
                        break
                    else:
                        state = self.FULLRED
            elif state == self.FULLRED:

                # Generating segmentation image
                _logger.info('Step %d, generating segmentation image', step)
                objmask, seeing_fwhm = self.create_mask(
                    sf_data, seeing_fwhm, step=step)
                step += 1
                # Update objects mask
                # For all images
                # FIXME:
                for frame in targetframes:
                    frame.objmask = name_object_mask(frame.baselabel, step)
                    _logger.info(
                        'Step %d, create object mask %s', step,  frame.objmask)
                    frame.objmask_data = objmask[frame.valid_region]
                    fits.writeto(
                        frame.objmask, frame.objmask_data, overwrite=True)

                if not target_is_sky:
                    # Empty object mask for sky frames
                    bogus_objmask = numpy.zeros(windowshape, dtype='int')

                    for frame in skyframes:
                        frame.objmask_data = bogus_objmask

                _logger.info('Step %d, superflat correction (SF)', step)

                # Compute scale factors (median)
                self.update_scale_factors(ri.obresult.frames, step)

                # Create superflat
                superflat = self.compute_superflat(skyframes, scaled_chan,
                                                   segmask=objmask, step=step)

                # Apply superflat
                self.figure_init(subpixshape)

                self.apply_superflat(
                    ri.obresult.frames, superflat, step=step, save=True)

                _logger.info('Step %d, advanced sky correction (SC)', step)
                self.compute_advanced_sky(targetframes, objmask,
                                          skyframes=skyframes,
                                          target_is_sky=target_is_sky,
                                          step=step)

                # Combining the images
                _logger.info("Step %d, Combining the images", step)
                # FIXME: only for science
                sf_data = self.combine_frames(
                    targetframes, ri.extinction, step=step)
                self.figures_after_combine(sf_data)

                if step >= niteration:
                    state = self.COMPLETE
            else:
                break

        if sf_data is None:
            raise RecipeError(
                'no combined image has been generated at step %d', state)

        hdu = fits.PrimaryHDU(sf_data[0])
        hdr = hdu.header
        hdr.update('NUMXVER', __version__, 'Numina package version')
        hdr.update('NUMRNAM', self.__class__.__name__, 'Numina recipe name')
        hdr.update('NUMRVER', self.__version__, 'Numina recipe version')

        hdr.update('FILENAME', 'result.fits')
        hdr.update('IMGTYP', 'TARGET', 'Image type')
        hdr.update('NUMTYP', 'TARGET', 'Data product type')

        varhdu = fits.ImageHDU(sf_data[1], name='VARIANCE')
        num = fits.ImageHDU(sf_data[2], name='MAP')

        result = fits.HDUList([hdu, varhdu, num])

        _logger.info("Final frame created")

        return DataFrame(result), SourcesCatalog()

    def compute_simple_sky(self, targetframes, skyframes,
                           maxsep=5, step=0, save=True):

        # build kdtree
        sarray = numpy.array([frame.mjd for frame in skyframes])
        # shape must be (n, 1)
        sarray = numpy.expand_dims(sarray, axis=1)

        # query
        tarray = numpy.array([frame.mjd for frame in targetframes])
        # shape must be (n, 1)
        tarray = numpy.expand_dims(tarray, axis=1)

        kdtree = KDTree(sarray)

        # 1 / minutes in a day
        MIN_TO_DAY = 0.000694444
        _dis, idxs = kdtree.query(tarray, k=1,
                                  distance_upper_bound=maxsep * MIN_TO_DAY)

        for tid, idss in enumerate(idxs):
            try:
                tf = targetframes[tid]
                sf = skyframes[idss]
                self.compute_simple_sky_for_frame(tf, sf, step=step, save=save)
            except IndexError:
                _logger.error(
                    'No sky image available for frame %s', tf.lastname)
                raise

    def compute_simple_sky_for_frame(self, frame, skyframe, step=0, save=True):
        _logger.info('Correcting sky in frame %s', frame.lastname)
        _logger.info('with sky computed from frame %s', skyframe.lastname)

        if hasattr(skyframe, 'median_sky'):
            sky = skyframe.median_sky
        else:

            with fits.open(skyframe.lastname, mode='readonly') as hdulist:
                data = hdulist['primary'].data
                valid = data[frame.valid_region]

                if skyframe.objmask_data is not None:
                    _logger.debug('object mask defined')
                    msk = frame.objmask_data
                    sky = numpy.median(valid[msk == 0])
                else:
                    _logger.debug('object mask empty')
                    sky = numpy.median(valid)

            _logger.debug('median sky value is %f', sky)
            skyframe.median_sky = sky

        dst = name_skysub_proc(frame.baselabel, step)
        prev = frame.lastname

        if save:
            shutil.copyfile(prev, dst)
        else:
            os.rename(prev, dst)

        frame.lastname = dst

        with fits.open(frame.lastname, mode='update') as hdulist:
            data = hdulist['primary'].data
            valid = data[frame.valid_region]
            valid -= sky

    def compute_advanced_sky(self, targetframes, objmask,
                             skyframes=None, target_is_sky=False,
                             maxsep=5.0,
                             nframes=10,
                             step=0, save=True):

        if target_is_sky:
            skyframes = targetframes
            # Each frame is its closets sky frame
            nframes += 1
        elif skyframes is None:
            raise ValueError('skyframes not defined')

        # build kdtree
        sarray = numpy.array([frame.mjd for frame in skyframes])
        # shape must be (n, 1)
        sarray = numpy.expand_dims(sarray, axis=1)

        # query
        tarray = numpy.array([frame.mjd for frame in targetframes])
        # shape must be (n, 1)
        tarray = numpy.expand_dims(tarray, axis=1)

        kdtree = KDTree(sarray)

        # 1 / minutes in a Julian day
        SCALE = 60.0
        # max_time_sep = ri.sky_images_sep_time / 1440.0
        _dis, idxs = kdtree.query(tarray, k=nframes,
                                  distance_upper_bound=maxsep * SCALE)

        nsky = len(sarray)

        for tid, idss in enumerate(idxs):
            try:
                tf = targetframes[tid]
                _logger.info('Step %d, SC: computing advanced sky for %s',
                             step, tf.baselabel)
                # filter(lambda x: x < nsky, idss)
                locskyframes = []
                for si in idss:
                    if tid == si:
                        # this sky frame its the current frame, reject
                        continue
                    if si < nsky:
                        _logger.debug('Step %d, SC: %s is a sky frame',
                                      step, skyframes[si].baselabel)
                        locskyframes.append(skyframes[si])
                self.compute_advanced_sky_for_frame(
                    tf, locskyframes, step=step, save=save)
            except IndexError:
                _logger.error(
                    'No sky image available for frame %s', tf.lastname)
                raise

    def compute_advanced_sky_for_frame(self, frame, skyframes,
                                       step=0, save=True):
        _logger.info('Correcting sky in frame %s', frame.lastname)
        _logger.info('with sky computed from frames')
        for i in skyframes:
            _logger.info('%s', i.flat_corrected)

        data = []
        scales = []
        masks = []
        # handle the FITS file to close it finally
        desc = []
        try:
            for i in skyframes:
                filename = i.flat_corrected
                hdulist = fits.open(filename, mode='readonly', memmap=True)

                data.append(hdulist['primary'].data[i.valid_region])
                desc.append(hdulist)
                scales.append(numpy.median(data[-1]))
                if i.objmask_data is not None:
                    masks.append(i.objmask_data)
                    _logger.debug('object mask is shared')
                elif i.objmask is not None:
                    hdulistmask = fits.open(
                        i.objmask, mode='readonly', memmap=True)
                    masks.append(hdulistmask['primary'].data)
                    desc.append(hdulistmask)
                    _logger.debug('object mask is particular')
                else:
                    _logger.warn('no object mask for %s', filename)

            _logger.debug('computing background with %d frames', len(data))
            sky, _, num = median(data, masks, scales=scales)

        finally:
            # Closing all FITS files
            for hdl in desc:
                hdl.close()

        if numpy.any(num == 0):
            # We have pixels without
            # sky background information
            _logger.warn('pixels without sky information when correcting %s',
                         frame.flat_corrected)
            binmask = num == 0
            # FIXME: during development, this is faster
            # sky[binmask] = sky[num != 0].mean()

            # To continue we interpolate over the patches
            fixpix2(sky, binmask, out=sky, iterations=1)
            name = name_skybackground(frame.baselabel, step)
            fits.writeto(name, sky, overwrite=True)
            name = name_skybackgroundmask(frame.baselabel, step)
            fits.writeto(name, binmask.astype('int16'), overwrite=True)

        dst = name_skysub_proc(frame.baselabel, step)
        prev = frame.lastname
        shutil.copyfile(prev, dst)
        frame.lastname = dst

        with fits.open(frame.lastname, mode='update') as hdulist:
            data = hdulist['primary'].data
            valid = data[frame.valid_region]
            valid -= sky

    def combine_frames(self, frames, extinction, out=None, step=0):
        _logger.debug('Step %d, opening sky-subtracted frames', step)

        def fits_open(name):
            '''Open FITS with memmap in readonly mode'''
            return fits.open(name, mode='readonly', memmap=True)

        frameslll = [fits_open(frame.lastname)
                     for frame in frames if frame.valid_target]
        _logger.debug('Step %d, opening mask frames', step)
        mskslll = [fits_open(frame.resized_mask)
                   for frame in frames if frame.valid_target]
        _logger.debug('Step %d, combining %d frames', step, len(frameslll))
        try:
            extinc = [pow(10, -0.4 * frame.airmass * extinction)
                      for frame in frames if frame.valid_target]
            data = [i['primary'].data for i in frameslll]
            masks = [i['primary'].data for i in mskslll]

            out = quantileclip(data, masks, scales=extinc,
                               dtype='float32', out=out, fclip=0.1)

            # saving the three extensions
            fits.writeto('result_i%0d.fits' % step, out[0], overwrite=True)
            fits.writeto('result_var_i%0d.fits' % step, out[1], overwrite=True)
            fits.writeto('result_npix_i%0d.fits' % step, out[2], overwrite=True)

            return out

        finally:
            _logger.debug('Step %d, closing sky-subtracted frames', step)
            for f in frameslll:
                f.close()
            _logger.debug('Step %d, closing mask frames', step)
            for f in mskslll:
                f.close()

    def apply_superflat(self, frames, flatdata, step=0, save=True):
        _logger.info("Step %d, SF: apply superflat", step)

        # Process all frames with the fitted flat
        # FIXME: not sure
        for frame in frames:
            self.correct_superflat(frame, flatdata, step=step, save=save)
        return frames

    def correct_superflat(self, frame, fitted, step=0, save=True):

        frame.flat_corrected = name_skyflat_proc(frame.baselabel, step)

        if save:
            shutil.copyfile(frame.resized_base, frame.flat_corrected)
        else:
            os.rename(frame.resized_base, frame.flat_corrected)

        _logger.info("Step %d, SF: apply superflat to frame %s",
                     step, frame.flat_corrected)
        with fits.open(frame.flat_corrected, mode='update') as hdulist:
            data = hdulist['primary'].data
            datar = data[frame.valid_region]
            data[frame.valid_region] = correct_flatfield(datar, fitted)

            frame.lastname = frame.flat_corrected

            # FIXME: plotting
            try:
                self.figure_image(data[frame.valid_region], frame)
            except ValueError:
                _logger.warning('Problem plotting %s', frame.lastname)

    def compute_superflat(self, frames, channels, segmask=None, step=0):
        _logger.info("Step %d, SF: combining the frames without offsets", step)
        try:
            filelist = []
            data = []
            masks = []
            for frame in frames:
                _logger.debug('Step %d, opening resized frame %s',
                              step, frame.resized_base)
                hdulist = fits.open(
                    frame.resized_base, memmap=True, mode='readonly')
                filelist.append(hdulist)
                data.append(hdulist['primary'].data[frame.valid_region])

            scales = [frame.median_scale for frame in frames]

            # FIXME: plotting
            self.figure_median_background(scales)

            if segmask is not None:
                masks = [segmask[frame.valid_region] for frame in frames]
            else:
                for frame in frames:
                    _logger.debug('Step %d, opening resized mask %s',
                                  step, frame.resized_mask)
                    hdulist = fits.open(
                        frame.resized_mask, memmap=True, mode='readonly')
                    filelist.append(hdulist)
                    masks.append(hdulist['primary'].data[frame.valid_region])

            _logger.debug('Step %d, combining %d frames', step, len(data))

            sf_data, _sf_var, sf_num = flatcombine(data, masks, scales=scales,
                                                   blank=1.0 / scales[0])
        finally:
            _logger.debug('Step %d, closing resized frames and mask', step)
            for fileh in filelist:
                fileh.close()

        # We interpolate holes by channel
        _logger.debug('Step %d, interpolating holes by channel', step)
        for channel in channels:
            mask = (sf_num[channel] == 0)
            if numpy.any(mask):
                fixpix2(sf_data[channel], mask, out=sf_data[channel])

        # Normalize, flat has mean = 1
        sf_data /= sf_data.mean()

        # Auxiliary data
        sfhdu = fits.PrimaryHDU(sf_data)
        sfhdu.writeto(name_skyflat('comb', step), overwrite=True)
        return sf_data

    def update_scale_factors(self, frames, step=0):
        _logger.info('Step %d, SF: computing scale factors', step)
        # FIXME: not sure
        for frame in frames:
            region = frame.valid_region
            data = fits.getdata(frame.resized_base)[region]
            mask = fits.getdata(frame.resized_mask)[region]
            # FIXME: while developing this ::10 is faster, remove later
            frame.median_scale = numpy.median(data[mask == 0][::10])
            _logger.debug('median value of %s is %f',
                          frame.resized_base, frame.median_scale)
        return frames

    def resize(self, frames, shape, offsetsp, finalshape, window=None,
               scale=1, step=0):
        _logger.info('Resizing frames and masks')
        for frame, rel_offset in zip(frames, offsetsp):
            if frame.valid_target:
                region, _ = subarray_match(finalshape, rel_offset, shape)
                # Valid region
                frame.valid_region = region
                # Relative offset
                frame.rel_offset = rel_offset
                # names of frame and mask
                framen, maskn = name_redimensioned_frames(
                    frame.baselabel, step)
                frame.resized_base = framen
                frame.resized_mask = maskn
                _logger.debug('%s, valid region is %s, relative offset is %s',
                              frame.label, custom_region_to_str(region),
                              rel_offset
                              )
                self.resize_frame_and_mask(
                    frame, finalshape, framen, maskn, window, scale)

        return frames

    def resize_frame_and_mask(self, frame, finalshape,
                              framen, maskn, window, scale):
        _logger.info('Resizing frame %s, window=%s, subpix=%i', frame.label,
                     custom_region_to_str(window), scale)
        hdul = frame.open()
        baseshape = hdul[0].data.shape

        # FIXME: Resize_fits saves the resized image in framen
        resize_fits(hdul, framen, finalshape, frame.valid_region,
                    window=window, scale=scale, dtype='float32')

        _logger.info('Resizing mask %s, subpix x%i', frame.label, scale)
        # We don't conserve the sum of the values of the frame here, just
        # expand the mask
        if frame.mask is None:
            self.logger.warning('BPM missing, use zeros instead')
            false_mask = numpy.zeros(baseshape, dtype='int16')
            hdum = fits.HDUList(fits.PrimaryHDU(false_mask))
            frame.mask = DataFrame(frame=hdum)
        resize_fits(frame.mask.open(), maskn, finalshape, frame.valid_region,
                    fill=1, window=window, scale=scale, conserve=False)

    def figure_init(self, shape):
        self._figure.clf()
        ax = self._figure.add_subplot(111)
        cmap = mpl.cm.get_cmap('gray')
        norm = mpl.colors.LogNorm()
        ax.imshow(numpy.ones(shape), cmap=cmap, norm=norm)
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        self._figure.canvas.draw()

    def figures_after_combine(self, sf_data, step=0):

        # FIXME, more plots
        def truncated(array, frac=0.1):
            '''Dirty truncated mean'''
            nf = int(array.size * frac)
            array.sort()
            new = array[nf:-nf]
            lnew = len(new)
            if lnew == 0:
                return 0, 0
            elif lnew == 1:
                return new.mean(), 0.0
            else:
                return new.mean(), new.std()

        ndata = sf_data[2].astype('int')
        data = sf_data[0]

        nimages = ndata.max()

        rnimage = list(six.moves.range(1, nimages + 1))
        rmean = rnimage[:]
        rstd = rnimage[:]

        for pix in rnimage:
            rmean[pix - 1], rstd[pix - 1] = truncated(data[ndata == pix])

        avg_rms = self.figure_check_combination(
            rnimage, rmean, rstd, step=step)

        # Fake sky error image
        self.figure_simple_image(ndata, title='Number of frames combined')

        # Create fake error frame
        mask = (ndata <= 0)
        ndata[mask] = 1
        fake = numpy.where(
            mask, 0.0, numpy.random.normal(avg_rms / numpy.sqrt(ndata)))
        ndata[mask] = 0
        self.figure_simple_image(fake, title='Fake sky error image')
        # store fake image
        fits.writeto('fake_sky_rms_i%0d.fits' % step, fake, overwrite=True)

    def figure_check_combination(self, rnimage, rmean, rstd, step=0):
        self._figure.clf()
        self._figure.subplots_adjust(hspace=0.001)

        ax1 = self._figure.add_subplot(3, 1, 1)
        pred = [rstd[-1] * math.sqrt(rnimage[-1] /
                                     float(npix)) for npix in rnimage]
        ax1.plot(rnimage, rstd, 'g*', rnimage, pred, 'y-')
        ax1.set_title("")
        ax1.set_ylabel('measured sky rms')

        ax2 = self._figure.add_subplot(3, 1, 2, sharex=ax1)
        pred = [val * math.sqrt(npix) for val, npix in zip(rstd, rnimage)]
        avg_rms = sum(pred) / len(pred)
        ax2.plot(
            rnimage, pred, 'r*', [rnimage[0], rnimage[-1]], [avg_rms, avg_rms])
        ax2.set_ylabel('scaled sky rms')

        ax3 = self._figure.add_subplot(3, 1, 3, sharex=ax1)
        ax3.plot(rnimage, rmean, 'b*')
        ax3.set_ylabel('mean sky')
        ax3.set_xlabel('number of frames per pixel')

        xticklabels = ax1.get_xticklabels() + ax2.get_xticklabels()
        mpl.artist.setp(xticklabels, visible=False)
        self._figure.canvas.draw()
        self._figure.savefig('figure-check-combination_i%01d.png' % step)
        return avg_rms

    def figure_simple_image(self, data, title=None):
        self._figure.clf()
        ax = self._figure.add_subplot(111)
        cmap = mpl.cm.get_cmap('gray')
        norm = mpl.colors.LogNorm()
        if title is not None:
            ax.set_title(title)

        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.imshow(data, cmap=cmap, norm=norm)
        self._figure.canvas.draw()

    def figure_median_background(self, scales, step=0):
        # FIXME: plotting
        self._figure.clf()
        ax = self._figure.add_subplot(1, 1, 1)
        ax.plot(scales, 'r*')
        ax.set_title("")
        ax.set_xlabel('Image number')
        ax.set_ylabel('Median')
        self._figure.canvas.draw()
        self._figure.savefig('figure-median-sky-background_i%01d.png' % step)

    def figure_image(self, thedata, image):
        ax = self._figure.gca()
        image_axes, = ax.get_images()
        image_axes.set_data(thedata)

        # Create normalizer object
        interval = PercentileInterval(50.)
        z1, z2 = interval.get_limits(thedata)
        norm = ImageNormalize(vmin=z1, vmax=z2, stretch=SqrtStretch())
        image_axes.set_clim(z1, z2)
        image_axes.set_norm(norm)
        clim = image_axes.get_clim()
        ax.set_title('%s, bg=%g fg=%g, linscale' %
                     (image.lastname, clim[0], clim[1]))
        self._figure.canvas.draw()

    def create_mask_single(self, frame, seeing_fwhm, step=0):
        #
        # remove_border = True

        # sextractor takes care of bad pixels
        sex = SExtractor()
        sex.config['CHECKIMAGE_TYPE'] = "SEGMENTATION"
        sex.config["CHECKIMAGE_NAME"] = name_object_mask(frame.baselabel, step)

        sex.config['VERBOSE_TYPE'] = 'QUIET'
        sex.config['PIXEL_SCALE'] = 1
        sex.config['BACK_TYPE'] = 'AUTO'

        if seeing_fwhm is not None and seeing_fwhm > 0:
            sex.config['SEEING_FWHM'] = seeing_fwhm * sex.config['PIXEL_SCALE']

        sex.config['PARAMETERS_LIST'].append('FLUX_BEST')
        sex.config['PARAMETERS_LIST'].append('X_IMAGE')
        sex.config['PARAMETERS_LIST'].append('Y_IMAGE')
        sex.config['PARAMETERS_LIST'].append('A_IMAGE')
        sex.config['PARAMETERS_LIST'].append('B_IMAGE')
        sex.config['PARAMETERS_LIST'].append('THETA_IMAGE')
        sex.config['PARAMETERS_LIST'].append('FWHM_IMAGE')
        sex.config['PARAMETERS_LIST'].append('CLASS_STAR')

        filename = frame.lastname

        # Lauch SExtractor on a FITS file
        sex.run(filename)

        # Plot objects
        # FIXME, plot sextractor objects on top of image
        patches = []
        fwhms = []
        nfirst = 0
        catalog_f = sopen(sex.config['CATALOG_NAME'])
        try:
            star = catalog_f.readline()
            while star:
                flags = star['FLAGS']
                # ignoring those objects with corrupted apertures
                if flags & sexcatalog.CORRUPTED_APER:
                    star = catalog_f.readline()
                    continue
                center = (star['X_IMAGE'], star['Y_IMAGE'])
                wd = 10 * star['A_IMAGE']
                hd = 10 * star['B_IMAGE']
                color = 'red'
                e = Ellipse(center, wd, hd, star['THETA_IMAGE'], color=color)
                patches.append(e)
                fwhms.append(star['FWHM_IMAGE'])
                nfirst += 1
                # FIXME Plot a ellipse
                star = catalog_f.readline()
        finally:
            catalog_f.close()

        p = PatchCollection(patches, alpha=0.4)
        ax = self._figure.gca()
        ax.add_collection(p)
        self._figure.canvas.draw()
        self._figure.savefig('figure-sky-segmentation-overlay_%01d.png' % step)

        self.figure_fwhm_histogram(fwhms, step=step)

        # mode with an histogram
        hist, edges = numpy.histogram(fwhms, 50)
        idx = hist.argmax()

        seeing_fwhm = 0.5 * (edges[idx] + edges[idx + 1])
        if seeing_fwhm <= 0:
            _logger.warning(
                'Seeing FHWM %f pixels is negative, reseting', seeing_fwhm)
            seeing_fwhm = None
        else:
            _logger.info('Seeing FHWM %f pixels (%f arcseconds)',
                         seeing_fwhm, seeing_fwhm * sex.config['PIXEL_SCALE'])
        name_segmask(step)
        _logger.info('Step %d, create object mask %s', step,  frame.objmask)
        frame.objmask = name_object_mask(frame.baselabel, step)
        frame.objmask_data = None
        return frame, seeing_fwhm

    def create_mask(self, sf_data, seeing_fwhm, step=0):
        # FIXME more plots
        self.figure_final_before_s(sf_data[0])

        #
        remove_border = True

        # sextractor takes care of bad pixels


        #if seeing_fwhm is not None and seeing_fwhm > 0:
        #    sex.config['SEEING_FWHM'] = seeing_fwhm * sex.config['PIXEL_SCALE']


        if remove_border:
            weigthmap = 'weights4rms.fits'

            # Create weight map, remove n pixs from either side
            # using a Hannig filter
            # npix = 90
            # w1 = npix
            # w2 = npix
            # wmap = numpy.ones_like(sf_data[0])

            # cos_win1 = numpy.hanning(2 * w1)
            # cos_win2 = numpy.hanning(2 * w2)

            # wmap[:,:w1] *= cos_win1[:w1]
            # wmap[:,-w1:] *= cos_win1[-w1:]
            # wmap[:w2,:] *= cos_win2[:w2, numpy.newaxis]
            # wmap[-w2:,:] *= cos_win2[-w2:, numpy.newaxis]

            # Take the number of combined images from the combined image
            wm = sf_data[2].copy()
            # Dont search objects where nimages < lower
            # FIXME: this is a magic number
            # We ignore objects in regions where we have less
            # than 10% of the images
            lower = sf_data[2].max() // 10
            border = (wm < lower)
            fits.writeto(weigthmap, border.astype('uint8'), overwrite=True)

            #sex.config['WEIGHT_TYPE'] = 'MAP_WEIGHT'
            # FIXME: this is a magic number
            # sex.config['WEIGHT_THRESH'] = 50
            #sex.config['WEIGHT_IMAGE'] = weigthmap
        else:
            border=None

        filename = 'result_i%0d.fits' % (step)

        # Lauch SExtractor on a FITS file
        #sex.run(filename)

        data_res = fits.getdata(filename)
        data_res = data_res.byteswap().newbyteorder()
        bkg = sep.Background(data_res)
        data_sub = data_res - bkg

        _logger.info('Runing source extraction tor in %s', filename)
        objects, objmask = sep.extract(data_sub, 1.5, err=bkg.globalrms,
                              mask=border, segmentation_map=True)
        fits.writeto(name_segmask(step), objmask, overwrite=True)

        # # Plot objects
        # # FIXME, plot sextractor objects on top of image
        # patches = []
        # fwhms = []
        # nfirst = 0
        # catalog_f = sopen(sex.config['CATALOG_NAME'])
        # try:
        #     star = catalog_f.readline()
        #     while star:
        #         flags = star['FLAGS']
        #         # ignoring those objects with corrupted apertures
        #         if flags & sexcatalog.CORRUPTED_APER:
        #             star = catalog_f.readline()
        #             continue
        #         center = (star['X_IMAGE'], star['Y_IMAGE'])
        #         wd = 10 * star['A_IMAGE']
        #         hd = 10 * star['B_IMAGE']
        #         color = 'red'
        #         e = Ellipse(center, wd, hd, star['THETA_IMAGE'], color=color)
        #         patches.append(e)
        #         fwhms.append(star['FWHM_IMAGE'])
        #         nfirst += 1
        #         # FIXME Plot a ellipse
        #         star = catalog_f.readline()
        # finally:
        #     catalog_f.close()
        #
        # p = PatchCollection(patches, alpha=0.4)
        # ax = self._figure.gca()
        # ax.add_collection(p)
        # self._figure.canvas.draw()
        # self._figure.savefig('figure-segmentation-overlay_%01d.png' % step)
        #
        # self.figure_fwhm_histogram(fwhms, step=step)
        #
        # # mode with an histogram
        # hist, edges = numpy.histogram(fwhms, 50)
        # idx = hist.argmax()
        #
        # seeing_fwhm = 0.5 * (edges[idx] + edges[idx + 1])
        # if seeing_fwhm <= 0:
        #     _logger.warning(
        #         'Seeing FHWM %f pixels is negative, reseting', seeing_fwhm)
        #     seeing_fwhm = None
        # else:
        #     _logger.info('Seeing FHWM %f pixels (%f arcseconds)',
        #                  seeing_fwhm, seeing_fwhm * sex.config['PIXEL_SCALE'])
        # objmask = fits.getdata(name_segmask(step))

        return objmask, seeing_fwhm

    def figure_final_before_s(self, data):
        self._figure.clf()
        ax = self._figure.add_subplot(111)
        cmap = mpl.cm.get_cmap('gray')
        # norm = mpl.colors.LogNorm()
        ax.set_title('Result image')
        ax.set_xlabel('X')
        ax.set_ylabel('Y')

        interval = PercentileInterval(50.)
        z1, z2 = interval.get_limits(data)
        norm = ImageNormalize(vmin=z1, vmax=z2, stretch=SqrtStretch())
        ax.imshow(data, cmap=cmap, clim=(z1, z2), norm=norm)
        self._figure.canvas.draw()

    def figure_fwhm_histogram(self, fwhms, step=0):
        self._figure.clf()
        ax = self._figure.add_subplot(111)
        ax.set_title('FWHM of objects')
        ax.hist(fwhms, 50, normed=1, facecolor='g', alpha=0.75)
        self._figure.canvas.draw()
        self._figure.savefig('figure-fwhm-histogram_i%01d.png' % step)



#
# Copyright 2008-2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


import logging

import numpy
import numina.util.node
import numina.processing as proc

import emirdrp.core as c

_logger = logging.getLogger(__name__)


def get_corrector_p(rinput, meta, ins, datamodel):
    key = 'master_bpm'
    info = meta.get(key)
    corrector_class = proc.BadPixelCorrector

    if info is not None:
        inputval = getattr(rinput, key)
        with inputval.open() as hdul:
            _logger.info('loading "%s"', key)
            _logger.debug('info: %s', info)
            corrector = corrector_class(
                hdul[0].data,
                datamodel=datamodel,
                calibid=datamodel.get_imgid(hdul)
            )
    else:
        _logger.info('"%s" not provided, ignored', key)
        corrector = numina.util.node.IdNode()

    return corrector


def get_corrector_b(rinput, meta, ins, datamodel):
    iinfo = meta['obresult']
    if iinfo:
        mode = iinfo[0]['readmode']
        if mode.lower() in c.EMIR_BIAS_MODES:
            use_bias = True
            _logger.info('readmode is %s, bias required', mode)
        else:
            use_bias = False
            _logger.info('readmode is %s, bias not required', mode)
    else:
        # raise ValueError('cannot gather images info')
        use_bias = False

    # Loading calibrations
    if use_bias:
        bias_info = meta['master_bias']
        with rinput.master_bias.open() as hdul:
            _logger.info('loading bias')
            _logger.debug('bias info: %s', bias_info)
            mbias = hdul[0].data
            bias_corrector = proc.BiasCorrector(
                mbias,
                datamodel=datamodel,
                calibid = datamodel.get_imgid(hdul)
            )
    else:
        _logger.info('ignoring bias')
        bias_corrector = numina.util.node.IdNode()

    return bias_corrector


def get_corrector_s(rinput, meta, ins, datamodel):
    sky_info = meta.get('master_sky')

    if sky_info is None:
        return numina.util.node.IdNode()
    else:
        with rinput.master_sky.open() as hdul:
            _logger.info('loading sky')
            _logger.debug('sky info: %s', sky_info)
            sky_corrector = proc.SkyCorrector(
                hdul[0].data,
                datamodel=datamodel,
                calibid=datamodel.get_imgid(hdul)
            )
        return sky_corrector


def get_corrector_f(rinput, meta, ins, datamodel):
    """Corrector for intensity flat"""
    from emirdrp.processing.flatfield import FlatFieldCorrector
    flat_info = meta['master_flat']
    with rinput.master_flat.open() as hdul:
        _logger.info('loading intensity flat')
        _logger.debug('flat info: %s', flat_info)
        mflat = hdul[0].data
        # Check NaN and Ceros
        mask1 = mflat < 0
        mask2 = ~numpy.isfinite(mflat)
        if numpy.any(mask1):
            _logger.warning('flat has %d values below 0', mask1.sum())
        if numpy.any(mask2):
            _logger.warning('flat has %d NaN', mask2.sum())
        flat_corrector = FlatFieldCorrector(mflat,
                                            datamodel=datamodel,
                                            calibid=datamodel.get_imgid(hdul))

    return flat_corrector


def get_corrector_sf(rinput, meta, ins, datamodel):
    """Corrector for spectral flat"""
    from emirdrp.processing.flatfield import FlatFieldCorrector
    flat_info = meta['master_flat']
    with rinput.master_flat.open() as hdul:
        _logger.info('loading spectral flat')
        _logger.debug('flat info: %s', flat_info)
        mflat = hdul[0].data
        # Check NaN and Ceros
        mask1 = mflat < 0
        mask2 = ~numpy.isfinite(mflat)
        if numpy.any(mask1):
            _logger.warning('flat has %d values below 0', mask1.sum())
        if numpy.any(mask2):
            _logger.warning('flat has %d NaN', mask2.sum())
        flat_corrector = FlatFieldCorrector(mflat,
                                            datamodel=datamodel,
                                            calibid=datamodel.get_imgid(hdul))

    return flat_corrector


def get_corrector_d(rinput, meta, ins, datamodel):
    key = 'master_dark'

    corrector = get_corrector_gen(rinput, datamodel, proc.DarkCorrector, key)
    return corrector


def get_corrector_gen(rinput, datamodel, CorrectorClass, key):
    value = getattr(rinput, key)
    req = rinput.stored()[key]
    if value is None:
        if req.optional:
            _logger.info('"%s" not provided, ignored', key)
            corrector = numina.util.node.IdNode()
            return corrector
        else:
            msg = '"{}" not provided, is required'.format(key)
            raise ValueError(msg)
    else:
        with value.open() as hdul:
            datac = hdul['primary'].data
            corrector = CorrectorClass(
                datac,
                calibid=datamodel.get_imgid(hdul),
                datamodel=datamodel
            )
        return corrector


def get_checker(rinput, meta, ins, datamodel):
    from emirdrp.processing.checkers import Checker
    return Checker()



#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Beam switched-nodded image mode recipe of EMIR"""


from numina.core import Parameter
from numina.core import RecipeInput
from numina.core.recipeinout import define_input, define_result
from numina.core import Result, DataFrameType, RecipeResult
from numina.core.requirements import ObservationResultRequirement

from emirdrp.products import SourcesCatalog
from emirdrp.requirements import MasterBadPixelMaskRequirement
from emirdrp.requirements import MasterBiasRequirement
from emirdrp.requirements import MasterDarkRequirement
from emirdrp.requirements import MasterIntensityFlatFieldRequirement
from emirdrp.requirements import Offsets_Requirement
from emirdrp.requirements import SkyImageSepTime_Requirement

from .shared import DirectImageCommon


class NBImageRecipeInput(RecipeInput):
    obresult = ObservationResultRequirement()
    master_bpm = MasterBadPixelMaskRequirement()
    master_bias = MasterBiasRequirement()
    master_dark = MasterDarkRequirement()
    master_flat = MasterIntensityFlatFieldRequirement()
    extinction = Parameter(0.0, 'Mean atmospheric extinction')
    sources = Parameter(
        [], 'List of x, y coordinates to measure FWHM', optional=True)
    offsets = Offsets_Requirement()
    sky_images = Parameter(5, 'Images used to estimate the background'
                           ' before and after current image')
    sky_images_sep_time = SkyImageSepTime_Requirement()
    check_photometry_levels = Parameter(
        [0.5, 0.8], 'Levels to check the flux of the objects')
    check_photometry_actions = Parameter(
        ['warn', 'warn', 'default'], 'Actions to take on images')


class NBImageRecipeResult(RecipeResult):
    frame = Result(DataFrameType)
    catalog = Result(SourcesCatalog)


@define_input(NBImageRecipeInput)
@define_result(NBImageRecipeResult)
class NBImageRecipe(DirectImageCommon):

    """
    The effect of recording a series of stare images, with the same
    acquisition parameters, and taken by pointing the TS in cycles
    between two, or more, sky positions. Displacements are larger
    than the EMIR FOV, so the images have no common area. Used
    for sky subtraction


    **Observing modes:**

        * Nodded/Beamswitched images

    """

    def run(self, ri):

        frame, catalog = self.process(ri,
                                      window=None, subpix=1,
                                      target_is_sky=False,
                                      stop_after=DirectImageCommon.FULLRED)

        return self.create_result(frame=frame, catalog=catalog)



#
# Copyright 2013-2016 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#


"""Corrector to remove flat field"""

from __future__ import division
#
import logging

import numpy
import numina.processing as proc


_logger = logging.getLogger(__name__)


class FlatFieldCorrector(proc.Corrector):
    """A Node that corrects a frame from flat-field."""

    def __init__(self, flatdata, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        self.update_variance = False

        super(FlatFieldCorrector, self).__init__(
            datamodel=datamodel,
            calibid=calibid,
            dtype=dtype)

        self.flatdata = flatdata
        self.flatdata[flatdata <= 0] = 1.0 # To avoid NaN
        self.flat_stats = flatdata.mean()

    def run(self, img):
        import datetime
        imgid = self.get_imgid(img)

        _logger.debug('correcting flat in %s', imgid)
        _logger.debug('flat mean is %f', self.flat_stats)

        data = self.datamodel.get_data(img)
        # data = array.correct_flatfield(data, self.flatdata, dtype=self.dtype)
        # Check if flatdata as 0
        mask1 = self.flatdata < 0
        if numpy.any(mask1):
            _logger.warning('flat has %d zeros', mask1.sum())
        mask2 = ~numpy.isfinite(data)
        if numpy.any(mask2):
            _logger.warning('image has %d NaN', mask2.sum())

        result = data / self.flatdata
        result = result.astype(self.dtype)

        # FIXME: not using datamodel
        img['primary'].data = result
        hdr = img['primary'].header
        hdr['NUM-FF'] = self.calibid
        hdr['history'] = 'Flat-field correction with {}'.format(self.calibid)
        hdr['history'] = 'Flat-field correction time {}'.format(datetime.datetime.utcnow().isoformat())
        hdr['history'] = 'Flat-field correction mean {}'.format(self.flat_stats)
        return img



#
# Copyright 2016-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Spectroscopy mode, Stare Spectra
"""

from astropy.io import fits
from datetime import datetime
import logging
import numpy as np
from skimage.feature import register_translation

from numina.core import Result
from numina.core import Requirement, Parameter
from numina.array.combine import sigmaclip
import numina.processing as proc
from numina.processing.combine import basic_processing_with_combination

import emirdrp.datamodel as datamodel
from emirdrp.core.recipe import EmirRecipe
from emirdrp.instrument.csu_configuration import CsuConfiguration
import emirdrp.products as prods
from emirdrp.products import RectWaveCoeff
from emirdrp.processing.wavecal.apply_rectwv_coeff import apply_rectwv_coeff
from emirdrp.processing.wavecal.median_slitlets_rectified \
    import median_slitlets_rectified
from emirdrp.processing.wavecal.rectwv_coeff_from_mos_library \
    import rectwv_coeff_from_mos_library
from emirdrp.processing.wavecal.rescale_array_z1z2 import rescale_array_to_z1z2
from emirdrp.processing.wavecal.retrieve_catlines import retrieve_catlines
from emirdrp.processing.wavecal.synthetic_lines_rawdata import \
    synthetic_lines_rawdata
from emirdrp.processing.wavecal.rectwv_coeff_to_ds9 import save_four_ds9
from emirdrp.processing.wavecal.rectwv_coeff_to_ds9 \
    import save_spectral_lines_ds9
from emirdrp.processing.wavecal.refine_rectwv_coeff import refine_rectwv_coeff
import emirdrp.requirements as reqs

from emirdrp.core import EMIR_MINIMUM_SLITLET_WIDTH_MM
from emirdrp.core import EMIR_MAXIMUM_SLITLET_WIDTH_MM
from emirdrp.core import EMIR_NBARS


class StareSpectraRecipe(EmirRecipe):
    """Process images in Stare spectra mode"""

    logger = logging.getLogger(__name__)

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterSpectralFlatFieldRequirement()
    master_sky = reqs.SpectralSkyRequirement(optional=True)

    stare = Result(prods.ProcessedMOS)

    def run(self, rinput):
        self.logger.info('starting stare spectra reduction')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow,
                                                    method=sigmaclip)
        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        # Update EXP to 0
        hdr['EXP'] = 0

        self.logger.info('end stare spectra reduction')
        result = self.create_result(stare=hdulist)
        return result


class StareSpectraWaveRecipe(EmirRecipe):
    """Process images in Stare spectra at the GTC.

    This recipe is intended to be used at GTC. The rectification
    and wavelength calibration can computed from a model if this
    model (master_rectwv) is provided as input.

    """

    logger = logging.getLogger(__name__)

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterSpectralFlatFieldRequirement()
    master_rectwv = reqs.MasterRectWaveRequirement(optional=True)
    master_sky = reqs.SpectralSkyRequirement(optional=True)

    reduced_image = Result(prods.ProcessedImage)
    reduced_mos = Result(prods.ProcessedMOS)

    def run(self, rinput):
        self.logger.info('starting reduction of stare spectra')

        self.logger.info(rinput.master_rectwv)

        # build object to proceed with bpm, bias, dark and flat
        flow = self.init_filters(rinput)

        # apply bpm, bias, dark and flat
        reduced_image = basic_processing_with_combination(rinput, flow,
                                                          method=sigmaclip)
        # update header with additional info
        hdr = reduced_image[0].header
        self.set_base_headers(hdr)

        # save intermediate image in work directory
        self.save_intermediate_img(reduced_image, 'reduced_image.fits')

        # rectification and wavelength calibration (if a model has
        # been provided)
        if rinput.master_rectwv:
            # RectWaveCoeff object with rectification and wavelength
            # calibration coefficients for the particular CSU configuration
            rectwv_coeff = rectwv_coeff_from_mos_library(
                reduced_image,
                rinput.master_rectwv
            )

            # apply rectification and wavelength calibration
            stare_image = apply_rectwv_coeff(
                reduced_image,
                rectwv_coeff
            )

            # save as JSON file in work directory
            self.save_structured_as_json(rectwv_coeff, 'rectwv_coeff.json')

            # ds9 region files (to be saved in the work directory)
            if self.intermediate_results:
                save_four_ds9(rectwv_coeff)
                save_spectral_lines_ds9(rectwv_coeff)

            # compute median spectra employing the useful region of the
            # rectified image
            if self.intermediate_results:
                for imode, outfile in enumerate(['median_spectra_full',
                                                 'median_spectra_slitlets',
                                                 'median_spectrum_slitlets']):
                    median_image = median_slitlets_rectified(
                        stare_image, mode=imode
                    )
                    self.save_intermediate_img(median_image, outfile + '.fits')

            # image_wl_calibrated = True

        else:

            stare_image = reduced_image

            self.logger.info('No wavelength calibration provided')
            grism_value = hdr.get('GRISM', 'unknown')
            self.logger.debug('GRISM is %s', grism_value)
            if grism_value.lower() == 'open':
                self.logger.debug('GRISM is %s, so this seems OK', grism_value)

            # image_wl_calibrated = False

        if rinput.master_sky:
            # Sky subtraction after rectification
            msky = rinput.master_sky.open()
            # Check if images have the same size.
            # if so, go ahead
            if msky[0].data.shape != stare_image[0].data.shape:
                self.logger.warning(
                    "sky and current image don't have the same shape")
            else:
                sky_corrector = proc.SkyCorrector(
                    msky[0].data,
                    datamodel=self.datamodel,
                    calibid=self.datamodel.get_imgid(msky)
                )

                stare_image = sky_corrector(stare_image)
        else:
            self.logger.info('No sky image provided')

        # save results in results directory
        self.logger.info('end reduction of stare spectra')
        result = self.create_result(reduced_image=reduced_image,
                                    reduced_mos=stare_image)
        return result

    def set_base_headers(self, hdr):
        newhdr = super(StareSpectraWaveRecipe, self).set_base_headers(hdr)
        # Update EXP to 0
        newhdr['EXP'] = 0
        return newhdr


class GenerateRectwvCoeff(EmirRecipe):
    """Process images in Stare spectra.

    This recipe generates a rectified and wavelength calibrated
    image after applying a model (master_rectwv). This calibration
    can be refined (using refine_wavecalib_mode != 0).

    """

    logger = logging.getLogger(__name__)

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterSpectralFlatFieldRequirement()
    master_rectwv = reqs.MasterRectWaveRequirement()
    refine_wavecalib_mode = Parameter(
        0,
        description='Apply wavelength calibration refinement',
        optional=True,
        choices=[0, 1, 2, 11, 12]
    )
    minimum_slitlet_width_mm = Parameter(
        float(EMIR_MINIMUM_SLITLET_WIDTH_MM),
        description='Minimum width (mm) for a valid slitlet',
        optional=True
    )
    maximum_slitlet_width_mm = Parameter(
        float(EMIR_MAXIMUM_SLITLET_WIDTH_MM),
        description='Maximum width (mm) for a valid slitlet',
        optional=True
    )
    global_integer_offsets_mode = Parameter(
        'fixed',
        description='Global integer offsets computation',
        choices=['auto', 'fixed']
    )
    global_integer_offset_x_pix = Parameter(
        0,
        description='Global integer offset (pixels) in wavelength direction',
        optional=True
    )
    global_integer_offset_y_pix = Parameter(
        0,
        description='Global integer offset (pixels) in spatial direction',
        optional=True
    )

    reduced_mos = Result(prods.ProcessedMOS)
    rectwv_coeff = Result(RectWaveCoeff)

    def run(self, rinput):
        self.logger.info('starting rect.+wavecal. reduction of stare spectra')

        self.logger.info(rinput.master_rectwv)
        self.logger.info(
            'Wavelength calibration refinement mode....: {}'.format(
                rinput.refine_wavecalib_mode))
        self.logger.info(
            'Minimum slitlet width (mm)................: {}'.format(
                rinput.minimum_slitlet_width_mm))
        self.logger.info(
            'Maximum slitlet width (mm)................: {}'.format(
                rinput.maximum_slitlet_width_mm))
        self.logger.info(
            'Global integer offsets mode...............: {}'.format(
                rinput.global_integer_offsets_mode
            )
        )
        self.logger.info(
            'Global integer offset X direction (pixels): {}'.format(
                rinput.global_integer_offset_x_pix))
        self.logger.info(
            'Global integer offset Y direction (pixels): {}'.format(
                rinput.global_integer_offset_y_pix))

        # build object to proceed with bpm, bias, dark and flat
        flow = self.init_filters(rinput)

        # apply bpm, bias, dark and flat
        reduced_image = basic_processing_with_combination(rinput, flow,
                                                          method=sigmaclip)
        # update header with additional info
        hdr = reduced_image[0].header
        self.set_base_headers(hdr)

        # save intermediate image in work directory
        self.save_intermediate_img(reduced_image, 'reduced_image.fits')

        # RectWaveCoeff object with rectification and wavelength
        # calibration coefficients for the particular CSU configuration
        rectwv_coeff = rectwv_coeff_from_mos_library(
            reduced_image,
            rinput.master_rectwv
        )

        # wavelength calibration refinement
        # 0 -> no refinement
        # 1 -> apply global offset to all the slitlets (using ARC lines)
        # 2 -> apply individual offset to each slitlet (using ARC lines)
        # 11 -> apply global offset to all the slitlets (using OH lines)
        # 12 -> apply individual offset to each slitlet (using OH lines)
        if rinput.refine_wavecalib_mode != 0:
            main_header = reduced_image[0].header
            mecs_header = datamodel.get_mecs_header(reduced_image)
            # determine useful slitlets
            csu_config = CsuConfiguration.define_from_header(mecs_header)
            # segregate slitlets
            list_useful_slitlets = csu_config.widths_in_range_mm(
                minwidth=rinput.minimum_slitlet_width_mm,
                maxwidth=rinput.maximum_slitlet_width_mm
            )
            # remove missing slitlets
            if len(rectwv_coeff.missing_slitlets) > 0:
                for iremove in rectwv_coeff.missing_slitlets:
                    if iremove in list_useful_slitlets:
                        list_useful_slitlets.remove(iremove)

            list_not_useful_slitlets = [i for i in
                                        list(range(1, EMIR_NBARS + 1))
                                        if i not in list_useful_slitlets]
            self.logger.info('list of useful slitlets: {}'.format(
                list_useful_slitlets))
            self.logger.info('list of unusable slitlets: {}'.format(
                list_not_useful_slitlets))

            # retrieve arc/OH lines
            catlines_all_wave, catlines_all_flux = retrieve_catlines(
                rinput.refine_wavecalib_mode,
                main_header['grism']
            )

            # global integer offsets
            if rinput.global_integer_offsets_mode == 'auto':
                if (rinput.global_integer_offset_x_pix != 0) or \
                        (rinput.global_integer_offset_y_pix != 0):
                    raise ValueError('Global integer offsets must be zero when'
                                     ' mode=auto')

                # ToDo: include additional airglow emission lines

                self.logger.info('computing synthetic image')
                # generate synthetic image
                synthetic_raw_data = synthetic_lines_rawdata(
                    catlines_all_wave,
                    catlines_all_flux,
                    list_useful_slitlets,
                    rectwv_coeff
                )
                synthetic_raw_header = main_header.copy()
                synthetic_raw_header['DATE-OBS'] = \
                    datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
                chistory = 'Synthetic image'
                synthetic_raw_header.add_history(chistory)
                hdu = fits.PrimaryHDU(synthetic_raw_data.astype('float32'),
                                      header=synthetic_raw_header)
                synthetic_raw_image = fits.HDUList([hdu])
                if self.intermediate_results:
                    self.save_intermediate_img(synthetic_raw_image,
                                               'synthetic_raw_image.fits')

                # cross-correlation to determine global integer offsets
                # (rescaling data arrays to [0, 1] before using skimage
                # function)
                data1_rs, coef1_rs = rescale_array_to_z1z2(
                    reduced_image[0].data, (0, 1)
                )
                data2_rs, coef2_rs = rescale_array_to_z1z2(
                    synthetic_raw_data, (0, 1)
                )
                shifts, error, diffphase = register_translation(
                    data1_rs, data2_rs, 100)
                self.logger.info('global_float_offset_x_pix..: {}'.format(
                    -shifts[1]
                ))
                self.logger.info('global_float_offset_y_pix..: {}'.format(
                    -shifts[0]
                ))
                rectwv_coeff.global_integer_offset_x_pix = \
                    -int(round(shifts[1]))
                rectwv_coeff.global_integer_offset_y_pix = \
                    -int(round(shifts[0]))
                self.logger.info('global_integer_offset_x_pix: {}'.format(
                    rectwv_coeff.global_integer_offset_x_pix
                ))
                self.logger.info('global_integer_offset_y_pix: {}'.format(
                    rectwv_coeff.global_integer_offset_y_pix
                ))
                if self.intermediate_results:
                    data_product = np.fft.fft2(data1_rs) * \
                                   np.fft.fft2(data2_rs).conj()
                    cc_image = np.fft.fftshift(np.fft.ifft2(data_product))
                    power = np.log10(cc_image.real)
                    hdu_power = fits.PrimaryHDU(power)
                    hdul_power = fits.HDUList([hdu_power])
                    hdul_power.writeto('power.fits', overwrite=True)
            else:
                rectwv_coeff.global_integer_offset_x_pix = \
                    rinput.global_integer_offset_x_pix
                rectwv_coeff.global_integer_offset_y_pix = \
                    rinput.global_integer_offset_y_pix

            # apply initial rectification and wavelength calibration
            reduced_mos = apply_rectwv_coeff(
                reduced_image,
                rectwv_coeff
            )

            self.logger.info(
                'Refining wavelength calibration (mode={})'.format(
                    rinput.refine_wavecalib_mode
                ))
            # refine RectWaveCoeff object
            rectwv_coeff, expected_catalog_lines = refine_rectwv_coeff(
                reduced_mos,
                rectwv_coeff,
                catlines_all_wave,
                catlines_all_flux,
                rinput.refine_wavecalib_mode,
                list_useful_slitlets,
                save_intermediate_results=self.intermediate_results
            )
            self.save_intermediate_img(expected_catalog_lines,
                                       'expected_catalog_lines.fits')

        # apply rectification and wavelength calibration
        reduced_mos = apply_rectwv_coeff(
            reduced_image,
            rectwv_coeff
        )

        # ds9 region files (to be saved in the work directory)
        if self.intermediate_results:
            save_four_ds9(rectwv_coeff)
            save_spectral_lines_ds9(rectwv_coeff)

        # compute median spectra employing the useful region of the
        # rectified image
        if self.intermediate_results:
            for imode, outfile in enumerate(['median_spectra_full',
                                             'median_spectra_slitlets',
                                             'median_spectrum_slitlets']):
                median_image = median_slitlets_rectified(
                    reduced_mos, mode=imode
                )
                self.save_intermediate_img(median_image, outfile + '.fits')

        # save results in results directory
        self.logger.info('end rect.+wavecal. reduction of stare spectra')
        result = self.create_result(reduced_mos=reduced_mos,
                                    rectwv_coeff=rectwv_coeff)
        return result

    def set_base_headers(self, hdr):
        newhdr = super(GenerateRectwvCoeff, self).set_base_headers(hdr)
        # Update EXP to 0
        newhdr['EXP'] = 0
        return newhdr


class StareSpectraRectwv(EmirRecipe):
    """Process images in Stare spectra mode applying wavelength calibration

    Note that in this case the wavelength calibration has already been
    determined.

    """
    logger = logging.getLogger(__name__)

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterSpectralFlatFieldRequirement()
    rectwv_coeff = reqs.RectWaveCoeffRequirement()

    reduced_mos = Result(prods.ProcessedMOS)

    def run(self, rinput):
        self.logger.info('applying existing rect.+wavecal. calibration of '
                         'stare spectra')

        # build object to proceed with bpm, bias, dark and flat
        flow = self.init_filters(rinput)

        # apply bpm, bias, dark and flat
        reduced_image = basic_processing_with_combination(rinput, flow,
                                                          method=sigmaclip)
        # update header with additional info
        hdr = reduced_image[0].header
        self.set_base_headers(hdr)

        # save intermediate image in work directory
        self.save_intermediate_img(reduced_image, 'reduced_image.fits')

        # apply rectification and wavelength calibration
        reduced_mos = apply_rectwv_coeff(
            reduced_image,
            rinput.rectwv_coeff
        )

        # ds9 region files (to be saved in the work directory)
        if self.intermediate_results:
            save_four_ds9(rinput.rectwv_coeff)
            save_spectral_lines_ds9(rinput.rectwv_coeff)

        # compute median spectra employing the useful region of the
        # rectified image
        if self.intermediate_results:
            for imode, outfile in enumerate(['median_spectra_full',
                                             'median_spectra_slitlets',
                                             'median_spectrum_slitlets']):
                median_image = median_slitlets_rectified(
                    reduced_mos, mode=imode
                )
                self.save_intermediate_img(median_image, outfile + '.fits')

        # save results in results directory
        self.logger.info('end rect.+wavecal. reduction of stare spectra')
        result = self.create_result(reduced_mos=reduced_mos)
        return result

    def set_base_headers(self, hdr):
        newhdr = super(StareSpectraRectwv, self).set_base_headers(hdr)
        # Update EXP to 0
        newhdr['EXP'] = 0
        return newhdr
